from __future__ import annotations

import os
from typing import TYPE_CHECKING

import click

from hatch.config.constants import AppEnvVars

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Create environments")
@click.argument("env_name", default="default")
@click.pass_obj
def create(app: Application, env_name: str):
    """Create environments."""
    app.ensure_environment_plugin_dependencies()

    environments = app.project.expand_environments(env_name)
    if not environments:
        app.abort(f"Environment `{env_name}` is not defined by project config")

    incompatible = {}
    for env in environments:
        environment = app.project.get_environment(env)
        if environment.exists():
            app.display_warning(f"Environment `{env}` already exists")
            continue

        try:
            environment.check_compatibility()
        except Exception as e:  # noqa: BLE001
            if env_name in app.project.config.matrices:
                incompatible[env] = str(e)
                continue

            app.abort(f"Environment `{env}` is incompatible: {e}")

        app.project.prepare_environment(environment, keep_env=bool(os.environ.get(AppEnvVars.KEEP_ENV)))

    if incompatible:
        num_incompatible = len(incompatible)
        app.display_warning(
            f"Skipped {num_incompatible} incompatible environment{'s' if num_incompatible > 1 else ''}:"
        )
        for env, reason in incompatible.items():
            app.display_warning(f"{env} -> {reason}")



from __future__ import annotations

from ast import literal_eval
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from hatch.utils.platform import Platform


class PythonInfo:
    def __init__(self, platform: Platform, executable: str = "python") -> None:
        self.platform = platform
        self.executable = executable

        self.__dep_check_data: dict[str, Any] | None = None
        self.__environment: dict[str, str] | None = None
        self.__sys_path: list[str] | None = None

    @property
    def dep_check_data(self) -> dict[str, Any]:
        if self.__dep_check_data is None:
            process = self.platform.check_command(
                [self.executable, "-W", "ignore", "-"], capture_output=True, input=DEP_CHECK_DATA_SCRIPT
            )

            self.__dep_check_data = literal_eval(process.stdout.strip().decode("utf-8"))

        return self.__dep_check_data

    @property
    def environment(self) -> dict[str, str]:
        if self.__environment is None:
            self.__environment = self.dep_check_data["environment"]

        return self.__environment

    @property
    def sys_path(self) -> list[str]:
        if self.__sys_path is None:
            self.__sys_path = self.dep_check_data["sys_path"]

        return self.__sys_path


# Keep support for Python 2 for a while:
# https://github.com/pypa/packaging/blob/20.9/packaging/markers.py#L267-L300
DEP_CHECK_DATA_SCRIPT = b"""\
import os
import platform
import sys

if hasattr(sys, 'implementation'):
    info = sys.implementation.version
    iver = '{0.major}.{0.minor}.{0.micro}'.format(info)
    kind = info.releaselevel
    if kind != 'final':
        iver += kind[0] + str(info.serial)
    implementation_name = sys.implementation.name
else:
    iver = '0'
    implementation_name = ''

environment = {
    'implementation_name': implementation_name,
    'implementation_version': iver,
    'os_name': os.name,
    'platform_machine': platform.machine(),
    'platform_python_implementation': platform.python_implementation(),
    'platform_release': platform.release(),
    'platform_system': platform.system(),
    'platform_version': platform.version(),
    'python_full_version': platform.python_version(),
    'python_version': '.'.join(platform.python_version_tuple()[:2]),
    'sys_platform': sys.platform,
}
sys_path = [path for path in sys.path if path]

print({'environment': environment, 'sys_path': sys_path})
"""



import os

FIELD_TO_PARSE = object()


class ConfigurationError(Exception):
    def __init__(self, *args, location):
        self.location = location
        super().__init__(*args)

    def __str__(self):
        return f"Error parsing config:\n{self.location}\n  {super().__str__()}"


def parse_config(obj):
    if isinstance(obj, LazilyParsedConfig):
        obj.parse_fields()
    elif isinstance(obj, list):
        for o in obj:
            parse_config(o)
    elif isinstance(obj, dict):
        for o in obj.values():
            parse_config(o)


class LazilyParsedConfig:
    def __init__(self, config: dict, steps: tuple = ()):
        self.raw_data = config
        self.steps = steps

    def parse_fields(self):
        for attribute in self.__dict__:
            _, prefix, name = attribute.partition("_field_")
            if prefix:
                parse_config(getattr(self, name))

    def raise_error(self, message, *, extra_steps=()):
        import inspect

        field = inspect.currentframe().f_back.f_code.co_name
        raise ConfigurationError(message, location=" -> ".join([*self.steps, field, *extra_steps]))


class RootConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_mode = FIELD_TO_PARSE
        self._field_project = FIELD_TO_PARSE
        self._field_shell = FIELD_TO_PARSE
        self._field_dirs = FIELD_TO_PARSE
        self._field_projects = FIELD_TO_PARSE
        self._field_publish = FIELD_TO_PARSE
        self._field_template = FIELD_TO_PARSE
        self._field_terminal = FIELD_TO_PARSE

    @property
    def mode(self):
        if self._field_mode is FIELD_TO_PARSE:
            if "mode" in self.raw_data:
                mode = self.raw_data["mode"]
                if not isinstance(mode, str):
                    self.raise_error("must be a string")

                valid_modes = ("aware", "local", "project")
                if mode not in valid_modes:
                    self.raise_error(f"must be one of: {', '.join(valid_modes)}")

                self._field_mode = mode
            else:
                self._field_mode = self.raw_data["mode"] = "local"

        return self._field_mode

    @mode.setter
    def mode(self, value):
        self.raw_data["mode"] = value
        self._field_mode = FIELD_TO_PARSE

    @property
    def project(self):
        if self._field_project is FIELD_TO_PARSE:
            if "project" in self.raw_data:
                project = self.raw_data["project"]
                if not isinstance(project, str):
                    self.raise_error("must be a string")

                self._field_project = project
            else:
                self._field_project = self.raw_data["project"] = ""

        return self._field_project

    @project.setter
    def project(self, value):
        self.raw_data["project"] = value
        self._field_project = FIELD_TO_PARSE

    @property
    def shell(self):
        if self._field_shell is FIELD_TO_PARSE:
            if "shell" in self.raw_data:
                shell = self.raw_data["shell"]
                if isinstance(shell, str):
                    self._field_shell = ShellConfig({"name": shell}, ("shell",))
                elif isinstance(shell, dict):
                    self._field_shell = ShellConfig(shell, ("shell",))
                else:
                    self.raise_error("must be a string or table")
            else:
                self.raw_data["shell"] = ""
                self._field_shell = ShellConfig({"name": ""}, ("shell",))

        return self._field_shell

    @shell.setter
    def shell(self, value):
        self.raw_data["shell"] = value
        self._field_shell = FIELD_TO_PARSE

    @property
    def dirs(self):
        if self._field_dirs is FIELD_TO_PARSE:
            if "dirs" in self.raw_data:
                dirs = self.raw_data["dirs"]
                if not isinstance(dirs, dict):
                    self.raise_error("must be a table")

                self._field_dirs = DirsConfig(dirs, ("dirs",))
            else:
                dirs = {}
                self.raw_data["dirs"] = dirs
                self._field_dirs = DirsConfig(dirs, ("dirs",))

        return self._field_dirs

    @dirs.setter
    def dirs(self, value):
        self.raw_data["dirs"] = value
        self._field_dirs = FIELD_TO_PARSE

    @property
    def projects(self):
        if self._field_projects is FIELD_TO_PARSE:
            if "projects" in self.raw_data:
                projects = self.raw_data["projects"]
                if not isinstance(projects, dict):
                    self.raise_error("must be a table")

                project_data = {}
                for name, data in projects.items():
                    if isinstance(data, str):
                        project_data[name] = ProjectConfig({"location": data}, ("projects", name))
                    elif isinstance(data, dict):
                        project_data[name] = ProjectConfig(data, ("projects", name))
                    else:
                        self.raise_error("must be a string or table", extra_steps=(name,))

                self._field_projects = project_data
            else:
                self._field_projects = self.raw_data["projects"] = {}

        return self._field_projects

    @projects.setter
    def projects(self, value):
        self.raw_data["projects"] = value
        self._field_projects = FIELD_TO_PARSE

    @property
    def publish(self):
        if self._field_publish is FIELD_TO_PARSE:
            if "publish" in self.raw_data:
                publish = self.raw_data["publish"]
                if not isinstance(publish, dict):
                    self.raise_error("must be a table")

                for name, data in publish.items():
                    if not isinstance(data, dict):
                        self.raise_error("must be a table", extra_steps=(name,))

                self._field_publish = publish
            else:
                self._field_publish = self.raw_data["publish"] = {"index": {"repo": "main"}}

        return self._field_publish

    @publish.setter
    def publish(self, value):
        self.raw_data["publish"] = value
        self._field_publish = FIELD_TO_PARSE

    @property
    def template(self):
        if self._field_template is FIELD_TO_PARSE:
            if "template" in self.raw_data:
                template = self.raw_data["template"]
                if not isinstance(template, dict):
                    self.raise_error("must be a table")

                self._field_template = TemplateConfig(template, ("template",))
            else:
                template = {}
                self.raw_data["template"] = template
                self._field_template = TemplateConfig(template, ("template",))

        return self._field_template

    @template.setter
    def template(self, value):
        self.raw_data["template"] = value
        self._field_template = FIELD_TO_PARSE

    @property
    def terminal(self):
        if self._field_terminal is FIELD_TO_PARSE:
            if "terminal" in self.raw_data:
                terminal = self.raw_data["terminal"]
                if not isinstance(terminal, dict):
                    self.raise_error("must be a table")

                self._field_terminal = TerminalConfig(terminal, ("terminal",))
            else:
                terminal = {}
                self.raw_data["terminal"] = terminal
                self._field_terminal = TerminalConfig(terminal, ("terminal",))

        return self._field_terminal

    @terminal.setter
    def terminal(self, value):
        self.raw_data["terminal"] = value
        self._field_terminal = FIELD_TO_PARSE


class ShellConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_name = FIELD_TO_PARSE
        self._field_path = FIELD_TO_PARSE
        self._field_args = FIELD_TO_PARSE

    @property
    def name(self):
        if self._field_name is FIELD_TO_PARSE:
            if "name" in self.raw_data:
                name = self.raw_data["name"]
                if not isinstance(name, str):
                    self.raise_error("must be a string")

                self._field_name = name
            else:
                self.raise_error("required field")

        return self._field_name

    @name.setter
    def name(self, value):
        self.raw_data["name"] = value
        self._field_name = FIELD_TO_PARSE

    @property
    def path(self):
        if self._field_path is FIELD_TO_PARSE:
            if "path" in self.raw_data:
                path = self.raw_data["path"]
                if not isinstance(path, str):
                    self.raise_error("must be a string")

                self._field_path = path
            else:
                self._field_path = self.raw_data["path"] = self.name

        return self._field_path

    @path.setter
    def path(self, value):
        self.raw_data["path"] = value
        self._field_path = FIELD_TO_PARSE

    @property
    def args(self):
        if self._field_args is FIELD_TO_PARSE:
            if "args" in self.raw_data:
                args = self.raw_data["args"]
                if not isinstance(args, list):
                    self.raise_error("must be an array")

                for i, entry in enumerate(args, 1):
                    if not isinstance(entry, str):
                        self.raise_error("must be a string", extra_steps=(str(i),))

                self._field_args = args
            else:
                self._field_args = self.raw_data["args"] = []

        return self._field_args

    @args.setter
    def args(self, value):
        self.raw_data["args"] = value
        self._field_args = FIELD_TO_PARSE


class DirsConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_project = FIELD_TO_PARSE
        self._field_env = FIELD_TO_PARSE
        self._field_python = FIELD_TO_PARSE
        self._field_data = FIELD_TO_PARSE
        self._field_cache = FIELD_TO_PARSE

    @property
    def project(self):
        if self._field_project is FIELD_TO_PARSE:
            if "project" in self.raw_data:
                project = self.raw_data["project"]
                if not isinstance(project, list):
                    self.raise_error("must be an array")

                for i, entry in enumerate(project, 1):
                    if not isinstance(entry, str):
                        self.raise_error("must be a string", extra_steps=(str(i),))

                self._field_project = project
            else:
                self._field_project = self.raw_data["project"] = []

        return self._field_project

    @project.setter
    def project(self, value):
        self.raw_data["project"] = value
        self._field_project = FIELD_TO_PARSE

    @property
    def env(self):
        if self._field_env is FIELD_TO_PARSE:
            if "env" in self.raw_data:
                env = self.raw_data["env"]
                if not isinstance(env, dict):
                    self.raise_error("must be a table")

                for key, value in env.items():
                    if not isinstance(value, str):
                        self.raise_error("must be a string", extra_steps=(key,))

                self._field_env = env
            else:
                self._field_env = self.raw_data["env"] = {}

        return self._field_env

    @env.setter
    def env(self, value):
        self.raw_data["env"] = value
        self._field_env = FIELD_TO_PARSE

    @property
    def python(self):
        if self._field_python is FIELD_TO_PARSE:
            if "python" in self.raw_data:
                python = self.raw_data["python"]
                if not isinstance(python, str):
                    self.raise_error("must be a string")

                self._field_python = python
            else:
                self._field_python = self.raw_data["python"] = "isolated"

        return self._field_python

    @python.setter
    def python(self, value):
        self.raw_data["python"] = value
        self._field_python = FIELD_TO_PARSE

    @property
    def data(self):
        if self._field_data is FIELD_TO_PARSE:
            if "data" in self.raw_data:
                data = self.raw_data["data"]
                if not isinstance(data, str):
                    self.raise_error("must be a string")

                self._field_data = data
            else:
                from platformdirs import user_data_dir

                self._field_data = self.raw_data["data"] = user_data_dir("hatch", appauthor=False)

        return self._field_data

    @data.setter
    def data(self, value):
        self.raw_data["data"] = value
        self._field_data = FIELD_TO_PARSE

    @property
    def cache(self):
        if self._field_cache is FIELD_TO_PARSE:
            if "cache" in self.raw_data:
                cache = self.raw_data["cache"]
                if not isinstance(cache, str):
                    self.raise_error("must be a string")

                self._field_cache = cache
            else:
                from platformdirs import user_cache_dir

                self._field_cache = self.raw_data["cache"] = user_cache_dir("hatch", appauthor=False)

        return self._field_cache

    @cache.setter
    def cache(self, value):
        self.raw_data["cache"] = value
        self._field_cache = FIELD_TO_PARSE


class ProjectConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_location = FIELD_TO_PARSE

    @property
    def location(self):
        if self._field_location is FIELD_TO_PARSE:
            if "location" in self.raw_data:
                location = self.raw_data["location"]
                if not isinstance(location, str):
                    self.raise_error("must be a string")

                self._field_location = location
            else:
                self.raise_error("required field")

        return self._field_location

    @location.setter
    def location(self, value):
        self.raw_data["location"] = value
        self._field_location = FIELD_TO_PARSE


class TemplateConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_name = FIELD_TO_PARSE
        self._field_email = FIELD_TO_PARSE
        self._field_licenses = FIELD_TO_PARSE
        self._field_plugins = FIELD_TO_PARSE

    @property
    def name(self):
        if self._field_name is FIELD_TO_PARSE:
            if "name" in self.raw_data:
                name = self.raw_data["name"]
                if not isinstance(name, str):
                    self.raise_error("must be a string")

                self._field_name = name
            else:
                name = os.environ.get("GIT_AUTHOR_NAME")
                if name is None:
                    import subprocess

                    try:
                        name = subprocess.check_output(
                            ["git", "config", "--get", "user.name"],  # noqa: S607
                            text=True,
                        ).strip()
                    except Exception:  # noqa: BLE001
                        name = "U.N. Owen"

                self._field_name = self.raw_data["name"] = name

        return self._field_name

    @name.setter
    def name(self, value):
        self.raw_data["name"] = value
        self._field_name = FIELD_TO_PARSE

    @property
    def email(self):
        if self._field_email is FIELD_TO_PARSE:
            if "email" in self.raw_data:
                email = self.raw_data["email"]
                if not isinstance(email, str):
                    self.raise_error("must be a string")

                self._field_email = email
            else:
                email = os.environ.get("GIT_AUTHOR_EMAIL")
                if email is None:
                    import subprocess

                    try:
                        email = subprocess.check_output(
                            ["git", "config", "--get", "user.email"],  # noqa: S607
                            text=True,
                        ).strip()
                    except Exception:  # noqa: BLE001
                        email = "void@some.where"

                self._field_email = self.raw_data["email"] = email

        return self._field_email

    @email.setter
    def email(self, value):
        self.raw_data["email"] = value
        self._field_email = FIELD_TO_PARSE

    @property
    def licenses(self):
        if self._field_licenses is FIELD_TO_PARSE:
            if "licenses" in self.raw_data:
                licenses = self.raw_data["licenses"]
                if not isinstance(licenses, dict):
                    self.raise_error("must be a table")

                self._field_licenses = LicensesConfig(licenses, (*self.steps, "licenses"))
            else:
                licenses = {}
                self.raw_data["licenses"] = licenses
                self._field_licenses = LicensesConfig(licenses, (*self.steps, "licenses"))

        return self._field_licenses

    @licenses.setter
    def licenses(self, value):
        self.raw_data["licenses"] = value
        self._field_licenses = FIELD_TO_PARSE

    @property
    def plugins(self):
        if self._field_plugins is FIELD_TO_PARSE:
            if "plugins" in self.raw_data:
                plugins = self.raw_data["plugins"]
                if not isinstance(plugins, dict):
                    self.raise_error("must be a table")

                for name, data in plugins.items():
                    if not isinstance(data, dict):
                        self.raise_error("must be a table", extra_steps=(name,))

                self._field_plugins = plugins
            else:
                self._field_plugins = self.raw_data["plugins"] = {
                    "default": {"tests": True, "ci": False, "src-layout": True}
                }

        return self._field_plugins

    @plugins.setter
    def plugins(self, value):
        self.raw_data["plugins"] = value
        self._field_plugins = FIELD_TO_PARSE


class LicensesConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_headers = FIELD_TO_PARSE
        self._field_default = FIELD_TO_PARSE

    @property
    def headers(self):
        if self._field_headers is FIELD_TO_PARSE:
            if "headers" in self.raw_data:
                headers = self.raw_data["headers"]
                if not isinstance(headers, bool):
                    self.raise_error("must be a boolean")

                self._field_headers = headers
            else:
                self._field_headers = self.raw_data["headers"] = True

        return self._field_headers

    @headers.setter
    def headers(self, value):
        self.raw_data["headers"] = value
        self._field_headers = FIELD_TO_PARSE

    @property
    def default(self):
        if self._field_default is FIELD_TO_PARSE:
            if "default" in self.raw_data:
                default = self.raw_data["default"]
                if not isinstance(default, list):
                    self.raise_error("must be an array")

                for i, entry in enumerate(default, 1):
                    if not isinstance(entry, str):
                        self.raise_error("must be a string", extra_steps=(str(i),))

                self._field_default = default
            else:
                self._field_default = self.raw_data["default"] = ["MIT"]

        return self._field_default

    @default.setter
    def default(self, value):
        self.raw_data["default"] = value
        self._field_default = FIELD_TO_PARSE


class TerminalConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_styles = FIELD_TO_PARSE

    @property
    def styles(self):
        if self._field_styles is FIELD_TO_PARSE:
            if "styles" in self.raw_data:
                styles = self.raw_data["styles"]
                if not isinstance(styles, dict):
                    self.raise_error("must be a table")

                self._field_styles = StylesConfig(styles, (*self.steps, "styles"))
            else:
                styles = {}
                self.raw_data["styles"] = styles
                self._field_styles = StylesConfig(styles, (*self.steps, "styles"))

        return self._field_styles

    @styles.setter
    def styles(self, value):
        self.raw_data["styles"] = value
        self._field_styles = FIELD_TO_PARSE


class StylesConfig(LazilyParsedConfig):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._field_info = FIELD_TO_PARSE
        self._field_success = FIELD_TO_PARSE
        self._field_error = FIELD_TO_PARSE
        self._field_warning = FIELD_TO_PARSE
        self._field_waiting = FIELD_TO_PARSE
        self._field_debug = FIELD_TO_PARSE
        self._field_spinner = FIELD_TO_PARSE

    @property
    def info(self):
        if self._field_info is FIELD_TO_PARSE:
            if "info" in self.raw_data:
                info = self.raw_data["info"]
                if not isinstance(info, str):
                    self.raise_error("must be a string")

                self._field_info = info
            else:
                self._field_info = self.raw_data["info"] = "bold"

        return self._field_info

    @info.setter
    def info(self, value):
        self.raw_data["info"] = value
        self._field_info = FIELD_TO_PARSE

    @property
    def success(self):
        if self._field_success is FIELD_TO_PARSE:
            if "success" in self.raw_data:
                success = self.raw_data["success"]
                if not isinstance(success, str):
                    self.raise_error("must be a string")

                self._field_success = success
            else:
                self._field_success = self.raw_data["success"] = "bold cyan"

        return self._field_success

    @success.setter
    def success(self, value):
        self.raw_data["success"] = value
        self._field_success = FIELD_TO_PARSE

    @property
    def error(self):
        if self._field_error is FIELD_TO_PARSE:
            if "error" in self.raw_data:
                error = self.raw_data["error"]
                if not isinstance(error, str):
                    self.raise_error("must be a string")

                self._field_error = error
            else:
                self._field_error = self.raw_data["error"] = "bold red"

        return self._field_error

    @error.setter
    def error(self, value):
        self.raw_data["error"] = value
        self._field_error = FIELD_TO_PARSE

    @property
    def warning(self):
        if self._field_warning is FIELD_TO_PARSE:
            if "warning" in self.raw_data:
                warning = self.raw_data["warning"]
                if not isinstance(warning, str):
                    self.raise_error("must be a string")

                self._field_warning = warning
            else:
                self._field_warning = self.raw_data["warning"] = "bold yellow"

        return self._field_warning

    @warning.setter
    def warning(self, value):
        self.raw_data["warning"] = value
        self._field_warning = FIELD_TO_PARSE

    @property
    def waiting(self):
        if self._field_waiting is FIELD_TO_PARSE:
            if "waiting" in self.raw_data:
                waiting = self.raw_data["waiting"]
                if not isinstance(waiting, str):
                    self.raise_error("must be a string")

                self._field_waiting = waiting
            else:
                self._field_waiting = self.raw_data["waiting"] = "bold magenta"

        return self._field_waiting

    @waiting.setter
    def waiting(self, value):
        self.raw_data["waiting"] = value
        self._field_waiting = FIELD_TO_PARSE

    @property
    def debug(self):
        if self._field_debug is FIELD_TO_PARSE:
            if "debug" in self.raw_data:
                debug = self.raw_data["debug"]
                if not isinstance(debug, str):
                    self.raise_error("must be a string")

                self._field_debug = debug
            else:
                self._field_debug = self.raw_data["debug"] = "bold"

        return self._field_debug

    @debug.setter
    def debug(self, value):
        self.raw_data["debug"] = value
        self._field_debug = FIELD_TO_PARSE

    @property
    def spinner(self):
        if self._field_spinner is FIELD_TO_PARSE:
            if "spinner" in self.raw_data:
                spinner = self.raw_data["spinner"]
                if not isinstance(spinner, str):
                    self.raise_error("must be a string")

                self._field_spinner = spinner
            else:
                self._field_spinner = self.raw_data["spinner"] = "simpleDotsScrolling"

        return self._field_spinner

    @spinner.setter
    def spinner(self, value):
        self.raw_data["spinner"] = value
        self._field_spinner = FIELD_TO_PARSE



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from hatch.publish.plugin.interface import PublisherInterface
from hatch.utils.fs import Path
from hatchling.metadata.utils import normalize_project_name

if TYPE_CHECKING:
    from collections.abc import Iterable


class IndexPublisher(PublisherInterface):
    PLUGIN_NAME = "index"

    def get_repos(self):
        global_plugin_config = self.plugin_config.copy()
        defined_repos = self.plugin_config.pop("repos", {})
        self.plugin_config.pop("repo", None)

        # Normalize type
        repos = {}
        for repo, data in defined_repos.items():
            if isinstance(data, str):
                repos[repo] = {"url": data}
            elif not isinstance(data, dict):
                self.app.abort(f"Hatch config field `publish.index.repos.{repo}` must be a string or a mapping")
            elif "url" not in data:
                self.app.abort(f"Hatch config field `publish.index.repos.{repo}` must define a `url` key")
            else:
                repos[repo] = data

        # Ensure PyPI correct
        for repo, url in (
            ("main", "https://upload.pypi.org/legacy/"),
            ("test", "https://test.pypi.org/legacy/"),
        ):
            repos.setdefault(repo, {})["url"] = url

        # Populate defaults
        for config in repos.values():
            for key, value in global_plugin_config.items():
                config.setdefault(key, value)

        return repos

    def publish(self, artifacts: list, options: dict):
        """
        https://warehouse.readthedocs.io/api-reference/legacy.html#upload-api
        """
        from collections import defaultdict

        from hatch.index.core import PackageIndex
        from hatch.index.publish import get_sdist_form_data, get_wheel_form_data
        from hatch.publish.auth import AuthenticationCredentials

        if not artifacts:
            from hatchling.builders.constants import DEFAULT_BUILD_DIRECTORY

            artifacts = [DEFAULT_BUILD_DIRECTORY]

        repo = options["repo"] if "repo" in options else self.plugin_config.get("repo", "main")
        repos = self.get_repos()
        repo_config: dict[str, str] = repos[repo] if repo in repos else {"url": repo}
        credentials = AuthenticationCredentials(
            app=self.app,
            cache_dir=self.cache_dir,
            options=options,
            repo=repo,
            repo_config=repo_config,
        )

        index = PackageIndex(
            repo_config["url"],
            user=credentials.username,
            auth=credentials.password,
            ca_cert=options.get("ca_cert", repo_config.get("ca-cert")),
            client_cert=options.get("client_cert", repo_config.get("client-cert")),
            client_key=options.get("client_key", repo_config.get("client-key")),
        )

        existing_artifacts: dict[str, set[str]] = {}

        # Use as an ordered set
        project_versions: dict[str, dict[str, None]] = defaultdict(dict)

        artifacts_found = False
        for artifact in recurse_artifacts(artifacts, self.root):
            if artifact.name.endswith(".whl"):
                data = get_wheel_form_data(artifact)
            elif artifact.name.endswith(".tar.gz"):
                data = get_sdist_form_data(artifact)
            else:
                continue

            artifacts_found = True

            for field in ("name", "version"):
                if field not in data:
                    self.app.abort(f"Missing required field `{field}` in artifact: {artifact}")

            try:
                displayed_path = str(artifact.relative_to(self.root))
            except ValueError:
                displayed_path = str(artifact)

            self.app.display_info(f"{displayed_path} ...", end=" ")

            project_name = normalize_project_name(data["name"])
            if project_name not in existing_artifacts:
                try:
                    response = index.get_simple_api(project_name)
                    response.raise_for_status()
                except Exception:  # no cov  # noqa: BLE001
                    existing_artifacts[project_name] = set()
                else:
                    existing_artifacts[project_name] = set(parse_artifacts(response.text))

            if artifact.name in existing_artifacts[project_name]:
                self.app.display_warning("already exists")
                continue

            try:
                index.upload_artifact(artifact, data)
            except Exception as e:  # noqa: BLE001
                self.app.display_error("failed")
                self.app.abort(f"Error uploading to repository: {index.repo} - {e}".replace(index.auth, "*****"))
            else:
                self.app.display_success("success")

                existing_artifacts[project_name].add(artifact.name)
                project_versions[project_name][data["version"]] = None

        if not options["initialize_auth"]:
            if not artifacts_found:
                self.app.abort("No artifacts found")
            elif not project_versions:
                self.app.abort(code=0)

        for project_name, versions in project_versions.items():
            self.app.display_info()
            self.app.display_mini_header(project_name)
            for version in versions:
                self.app.display_info(str(index.urls.project.child(project_name, version, "").to_iri()))

        credentials.write_updated_data()


def recurse_artifacts(artifacts: list, root) -> Iterable[Path]:
    for raw_artifact in artifacts:
        artifact = Path(raw_artifact)
        if not artifact.is_absolute():
            artifact = root / artifact

        if artifact.is_file():
            yield artifact
        elif artifact.is_dir():
            yield from artifact.iterdir()


def parse_artifacts(artifact_payload):
    for match in re.finditer(r"<a [^>]+>([^<]+)</a>", artifact_payload):
        yield match.group(1)



from __future__ import annotations

from typing import cast

from hatch.config.model import RootConfig
from hatch.utils.fs import Path
from hatch.utils.toml import load_toml_data


class ConfigFile:
    def __init__(self, path: Path | None = None):
        self._path: Path | None = path
        self.model = cast(RootConfig, None)

    @property
    def path(self):
        if self._path is None:
            self._path = self.get_default_location()

        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    def save(self, content=None):
        import tomli_w

        if not content:
            content = tomli_w.dumps(self.model.raw_data)

        self.path.ensure_parent_dir_exists()
        self.path.write_atomic(content, "w", encoding="utf-8")

    def load(self):
        self.model = RootConfig(load_toml_data(self.read()))

    def read(self) -> str:
        return self.path.read_text("utf-8")

    def read_scrubbed(self) -> str:
        import tomli_w

        config = RootConfig(load_toml_data(self.read()))
        config.raw_data.pop("publish", None)
        return tomli_w.dumps(config.raw_data)

    def restore(self):
        import tomli_w

        config = RootConfig({})
        config.parse_fields()

        content = tomli_w.dumps(config.raw_data)
        self.save(content)

        self.model = config

    def update(self):  # no cov
        self.model.parse_fields()
        self.save()

    @classmethod
    def get_default_location(cls) -> Path:
        from platformdirs import user_config_dir

        return Path(user_config_dir("hatch", appauthor=False)) / "config.toml"



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from configparser import ConfigParser

    from hatch.utils.fs import Path


class PatchedCoverageConfig:
    def __init__(self, project_root: Path, data_dir: Path) -> None:
        self.project_root = project_root
        self.data_dir = data_dir

    @cached_property
    def user_config_path(self) -> Path:
        # https://coverage.readthedocs.io/en/7.4.4/config.html#sample-file
        return (
            dedicated_coverage_file
            if (dedicated_coverage_file := self.project_root.joinpath(".coveragerc")).is_file()
            else self.project_root.joinpath("pyproject.toml")
        )

    @cached_property
    def internal_config_path(self) -> Path:
        return self.data_dir / "coverage" / self.project_root.id / self.user_config_path.name

    def write_config_file(self) -> None:
        self.internal_config_path.parent.ensure_dir_exists()
        if self.internal_config_path.name == ".coveragerc":
            from configparser import ConfigParser

            cfg = ConfigParser()
            cfg.read(str(self.user_config_path))

            if "run" not in cfg:
                cfg["run"] = {"parallel": "true"}
                self._write_ini(cfg)
                return

            cfg["run"]["parallel"] = "true"
            self._write_ini(cfg)
        else:
            import tomli_w

            from hatch.utils.toml import load_toml_data

            project_data = load_toml_data(self.user_config_path.read_text())
            project_data.setdefault("tool", {}).setdefault("coverage", {}).setdefault("run", {})["parallel"] = True
            self.internal_config_path.write_text(tomli_w.dumps(project_data))

    def _write_ini(self, cfg: ConfigParser) -> None:
        with self.internal_config_path.open("w", encoding="utf-8") as f:
            cfg.write(f)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application
    from hatch.env.plugin.interface import EnvironmentInterface


@click.command(short_help="Run tests", context_settings={"ignore_unknown_options": True})
@click.argument("args", nargs=-1)
@click.option("--randomize", "-r", is_flag=True, help="Randomize the order of test execution")
@click.option("--parallel", "-p", is_flag=True, help="Parallelize test execution")
@click.option("--retries", type=int, help="Number of times to retry failed tests")
@click.option("--retry-delay", type=float, help="Seconds to wait between retries")
@click.option("--cover", "-c", is_flag=True, help="Measure code coverage")
@click.option("--cover-quiet", is_flag=True, help="Disable coverage reporting after tests, implicitly enabling --cover")
@click.option(
    "--cover-xml",
    is_flag=True,
    help="Generate XML coverage report after tests, implicitly enabling --cover",
)
@click.option(
    "--cover-xml-output",
    type=click.Path(),
    help="Path for the XML coverage report file",
)
@click.option("--all", "-a", "test_all", is_flag=True, help="Test all environments in the matrix")
@click.option("--python", "-py", help="The Python versions to test, equivalent to: -i py=...")
@click.option("--include", "-i", "included_variable_specs", multiple=True, help="The matrix variables to include")
@click.option("--exclude", "-x", "excluded_variable_specs", multiple=True, help="The matrix variables to exclude")
@click.option("--show", "-s", is_flag=True, help="Show information about environments in the matrix")
@click.pass_context
def test(
    ctx: click.Context,
    *,
    args: tuple[str, ...],
    randomize: bool,
    parallel: bool,
    retries: int | None,
    retry_delay: float | None,
    cover: bool,
    cover_quiet: bool,
    cover_xml: bool,
    cover_xml_output: str | None,
    test_all: bool,
    python: str | None,
    included_variable_specs: tuple[str, ...],
    excluded_variable_specs: tuple[str, ...],
    show: bool,
):
    """Run tests using the `hatch-test` environment matrix.

    If no filtering options are selected, then tests will be run in the first compatible environment
    found in the matrix with priority given to those matching the current interpreter.

    The `-i`/`--include` and `-x`/`--exclude` options may be used to include or exclude certain
    variables, optionally followed by specific comma-separated values, and may be selected multiple
    times. For example, if you have the following configuration:

    \b
    ```toml config-example
    [[tool.hatch.envs.hatch-test.matrix]]
    python = ["3.9", "3.10"]
    version = ["42", "3.14", "9000"]
    ```

    then running:

    \b
    ```
    hatch test -i py=3.10 -x version=9000
    ```

    would run tests in the environments `hatch-test.py3.10-42` and `hatch-test.py3.10-3.14`.

    The `-py`/`--python` option is a shortcut for specifying the inclusion `-i py=...`.

    \b
    !!! note
        The inclusion option is treated as an intersection while the exclusion option is treated as a
        union i.e. an environment must match all of the included variables to be selected while matching
        any of the excluded variables will prevent selection.

    The test environment comes with default dependencies (pytest, coverage, etc.). Use the
    `-s`/`--show` option to see the full resolved configuration for each environment.
    To customize dependencies or other settings, see the testing configuration docs:
    https://hatch.pypa.io/latest/config/internal/testing/
    """
    app: Application = ctx.obj

    if show:
        import os

        from hatch.cli.env.show import show as show_env

        ctx.invoke(
            show_env,
            internal=True,
            hide_titles=True,
            force_ascii=os.environ.get("HATCH_SELF_TESTING") == "true",
            envs=ctx.obj.project.config.internal_matrices["hatch-test"]["envs"] if app.verbose else ["hatch-test"],
        )
        return

    if cover_quiet:
        cover = True

    if cover_xml:
        cover = True

    import sys

    from hatch.cli.test.core import PatchedCoverageConfig
    from hatch.utils.runner import parse_matrix_variables, select_environments

    if python is not None:
        included_variable_specs = (f"py={python}", *included_variable_specs)

    try:
        included_variables = parse_matrix_variables(included_variable_specs)
    except ValueError as e:
        app.abort(f"Duplicate included variable: {e}")

    try:
        excluded_variables = parse_matrix_variables(excluded_variable_specs)
    except ValueError as e:
        app.abort(f"Duplicate excluded variable: {e}")

    if test_all and (included_variables or excluded_variables):
        app.abort("The --all option cannot be used with the --include or --exclude options.")

    if retries is None and retry_delay is not None:
        app.abort("The --retry-delay option requires the --retries option to be set as well.")

    app.ensure_environment_plugin_dependencies()

    test_envs = app.project.config.internal_matrices["hatch-test"]["envs"]
    selected_envs: list[str] = []
    multiple_possible = True
    if test_all:
        selected_envs.extend(test_envs)
    elif included_variables or excluded_variables:
        selected_envs.extend(select_environments(test_envs, included_variables, excluded_variables))
    else:
        multiple_possible = False

        # Prioritize candidates that seems to match the running interpreter
        current_version = ".".join(map(str, sys.version_info[:2]))
        candidate_names: list[str] = list(test_envs)
        candidate_names.sort(key=lambda name: test_envs[name].get("python") != current_version)

        candidate_envs: list[EnvironmentInterface] = []
        for candidate_name in candidate_names:
            environment = app.project.get_environment(candidate_name)
            if environment.exists():
                selected_envs.append(candidate_name)
                break

            candidate_envs.append(environment)

        if not selected_envs:
            # If none of the candidates exist, then do a check for compatibility
            for candidate_env in candidate_envs:
                try:
                    candidate_env.check_compatibility()
                except Exception:  # noqa: BLE001, S110
                    pass
                else:
                    selected_envs.append(candidate_env.name)
                    break
            else:
                app.abort(f"No compatible environments found: {candidate_envs}")

    test_script = "run-cov" if cover else "run"
    patched_coverage = PatchedCoverageConfig(app.project.location, app.data_dir / ".config")
    coverage_config_file = str(patched_coverage.internal_config_path)
    if cover:
        patched_coverage.write_config_file()

    for context in app.runner_context(selected_envs, ignore_compat=multiple_possible, display_header=multiple_possible):
        internal_arguments: list[str] = list(context.env.config.get("extra-args", []))

        if not context.env.config.get("randomize", randomize):
            internal_arguments.extend(["-p", "no:randomly"])

        if context.env.config.get("parallel", parallel):
            internal_arguments.extend(["-n", "logical"])

        if (num_retries := context.env.config.get("retries", retries)) is not None:
            if "-r" not in args:
                internal_arguments.extend(["-r", "aR"])

            internal_arguments.extend(["--reruns", str(num_retries)])

        if (seconds_delay := context.env.config.get("retry-delay", retry_delay)) is not None:
            internal_arguments.extend(["--reruns-delay", str(seconds_delay)])

        internal_args = context.env.join_command_args(internal_arguments)
        if internal_args:
            # Add an extra space if required
            internal_args = f" {internal_args}"

        arguments: list[str] = []
        if args:
            arguments.extend(args)
        else:
            arguments.extend(context.env.config.get("default-args", ["tests"]))

        context.add_shell_command([test_script, *arguments])
        context.env_vars["HATCH_TEST_ARGS"] = internal_args
        if cover:
            context.env_vars["COVERAGE_RCFILE"] = coverage_config_file
            context.env_vars["COVERAGE_PROCESS_START"] = coverage_config_file

    if cover:
        for context in app.runner_context([selected_envs[0]]):
            context.add_shell_command("cov-combine")

        if not cover_quiet:
            for context in app.runner_context([selected_envs[0]]):
                context.add_shell_command("cov-report")

        if cover_xml:
            xml_command = f"coverage xml -o {cover_xml_output}" if cover_xml_output else "coverage xml"

            for context in app.runner_context([selected_envs[0]]):
                context.add_shell_command(xml_command)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Remove Python distributions")
@click.argument("names", required=True, nargs=-1)
@click.option("--dir", "-d", "directory", help="The directory in which distributions reside")
@click.pass_obj
def remove(app: Application, *, names: tuple[str, ...], directory: str | None):
    """
    Remove Python distributions.

    You may select `all` to remove all installed distributions:

    \b
    ```
    hatch python remove all
    ```
    """
    manager = app.get_python_manager(directory)
    installed = manager.get_installed()
    selection = tuple(installed) if "all" in names else names
    for name in selection:
        if name not in installed:
            app.display_warning(f"Distribution is not installed: {name}")
            continue

        dist = installed[name]
        with app.status(f"Removing {name}"):
            manager.remove(dist)



from __future__ import annotations

import re
from collections import defaultdict
from collections.abc import Generator
from contextlib import contextmanager
from functools import cached_property
from typing import TYPE_CHECKING, Any, cast

from hatch.project.env import EnvironmentMetadata
from hatch.utils.fs import Path
from hatch.utils.runner import ExecutionContext

if TYPE_CHECKING:
    from collections.abc import Generator

    from hatch.cli.application import Application
    from hatch.config.model import RootConfig
    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.project.frontend.core import BuildFrontend


class Project:
    def __init__(self, path: Path, *, name: str | None = None, config=None, locate: bool = True):
        self._path = path

        # From app config
        self.chosen_name = name

        # Lazily attach the current app
        self.__app: Application | None = None

        # Location of pyproject.toml
        self._project_file_path: Path | None = None

        self._root_searched = False
        self._root: Path | None = None
        self._raw_config = config
        self._plugin_manager = None
        self._metadata = None
        self._config = None

        self._explicit_path: Path | None = None if locate else path
        self.current_member_path: Path | None = None

    @property
    def plugin_manager(self):
        if self._plugin_manager is None:
            from hatch.plugin.manager import PluginManager

            self._plugin_manager = PluginManager()

        return self._plugin_manager

    @property
    def config(self):
        if self._config is None:
            from hatch.project.config import ProjectConfig

            self._config = ProjectConfig(self.location, self.metadata.hatch.config, self.plugin_manager)

        return self._config

    @property
    def root(self) -> Path | None:
        if not self._root_searched:
            self._root = self.find_project_root()
            self._root_searched = True

        return self._root

    @property
    def location(self) -> Path:
        return self._explicit_path or self.root or self._path

    def set_path(self, path: Path) -> None:
        self._explicit_path = path

    def set_app(self, app: Application) -> None:
        self.__app = app

    @cached_property
    def app(self) -> Application:
        if self.__app is None:  # no cov
            message = "The application has not been set"
            raise RuntimeError(message)

        from hatch.cli.application import Application

        return cast(Application, self.__app)

    @cached_property
    def build_env(self) -> EnvironmentInterface:
        # Prevent the default environment from being used as a builder environment
        environment = self.get_environment("hatch-build" if self.app.env == "default" else self.app.env)
        if not environment.builder:
            self.app.abort(f"Environment `{environment.name}` is not a builder environment")

        return environment

    @cached_property
    def build_frontend(self) -> BuildFrontend:
        from hatch.project.frontend.core import BuildFrontend

        return BuildFrontend(self, self.build_env)

    @cached_property
    def env_metadata(self) -> EnvironmentMetadata:
        return EnvironmentMetadata(self.app.data_dir / "env" / ".metadata", self.location)

    @cached_property
    def dependency_groups(self) -> dict[str, Any]:
        """
        https://peps.python.org/pep-0735/
        """
        from hatch.utils.metadata import normalize_project_name

        dependency_groups = self.raw_config.get("dependency-groups", {})

        if not isinstance(dependency_groups, dict):
            message = "Field `dependency-groups` must be a table"
            raise TypeError(message)

        original_names = defaultdict(list)
        normalized_groups = {}

        for group_name, value in dependency_groups.items():
            normed_group_name = normalize_project_name(group_name)
            original_names[normed_group_name].append(group_name)
            normalized_groups[normed_group_name] = value

        errors = []
        for normed_name, names in original_names.items():
            if len(names) > 1:
                errors.append(f"{normed_name} ({', '.join(names)})")
        if errors:
            msg = f"Field `dependency-groups` contains duplicate names: {', '.join(errors)}"
            raise ValueError(msg)

        return normalized_groups

    def get_environment(self, env_name: str | None = None) -> EnvironmentInterface:
        if env_name is None:
            env_name = self.app.env

        if env_name in self.config.internal_envs:
            config = self.config.internal_envs[env_name]
        elif env_name in self.config.envs:
            config = self.config.envs[env_name]
        else:
            self.app.abort(f"Unknown environment: {env_name}")

        environment_type = config["type"]
        environment_class = self.plugin_manager.environment.get(environment_type)
        if environment_class is None:
            self.app.abort(f"Environment `{env_name}` has unknown type: {environment_type}")

        from hatch.env.internal import is_isolated_environment

        if self.location.is_file():
            data_directory = isolated_data_directory = self.app.data_dir / "env" / environment_type / ".scripts"
        elif is_isolated_environment(env_name, config):
            data_directory = isolated_data_directory = self.app.data_dir / "env" / ".internal" / env_name
        else:
            data_directory = self.app.get_env_directory(environment_type)
            isolated_data_directory = self.app.data_dir / "env" / environment_type

        self.config.finalize_env_overrides(environment_class.get_option_types())

        return environment_class(
            self.location,
            self.metadata,
            env_name,
            config,
            self.config.matrix_variables.get(env_name, {}),
            data_directory,
            isolated_data_directory,
            self.app.platform,
            self.app.verbosity,
            self.app,
        )

    @staticmethod
    @contextmanager
    def managed_environment(
        environment: EnvironmentInterface, *, keep_env: bool = False
    ) -> Generator[EnvironmentInterface, None, None]:
        """Context manager that removes environment on error unless keep_env is True."""
        try:
            yield environment
        except Exception:
            if not keep_env and environment.exists():
                environment.remove()
            raise

    # Ensure that this method is clearly written since it is
    # used for documenting the life cycle of environments.
    def prepare_environment(self, environment: EnvironmentInterface, *, keep_env: bool):
        if not environment.exists():
            with self.managed_environment(environment, keep_env=keep_env):
                self.env_metadata.reset(environment)

                with environment.app_status_creation():
                    environment.create()

                if not environment.skip_install:
                    if environment.pre_install_commands:
                        with environment.app_status_pre_installation():
                            self.app.run_shell_commands(
                                ExecutionContext(
                                    environment,
                                    shell_commands=environment.pre_install_commands,
                                    source="pre-install",
                                    show_code_on_error=True,
                                )
                            )

                    with environment.app_status_project_installation():
                        if environment.dev_mode:
                            environment.install_project_dev_mode()
                        else:
                            environment.install_project()

                    if environment.post_install_commands:
                        with environment.app_status_post_installation():
                            self.app.run_shell_commands(
                                ExecutionContext(
                                    environment,
                                    shell_commands=environment.post_install_commands,
                                    source="post-install",
                                    show_code_on_error=True,
                                )
                            )

        with environment.app_status_dependency_state_check():
            new_dep_hash = environment.dependency_hash()

        current_dep_hash = self.env_metadata.dependency_hash(environment)

        from hatch.env.lock import environment_has_lock_inputs, generate_lockfile, resolve_lockfile_path

        if environment.locked and environment_has_lock_inputs(environment):
            lockfile_path = resolve_lockfile_path(environment)
            if not lockfile_path.is_file() or new_dep_hash != current_dep_hash:
                with self.app.status(f"Locking environment: {environment.name}"):
                    generate_lockfile(environment, lockfile_path)

        if new_dep_hash != current_dep_hash:
            with environment.app_status_dependency_installation_check():
                dependencies_in_sync = environment.dependencies_in_sync()

            if not dependencies_in_sync:
                with environment.app_status_dependency_synchronization():
                    environment.sync_dependencies()
                    new_dep_hash = environment.dependency_hash()

            self.env_metadata.update_dependency_hash(environment, new_dep_hash)

    def prepare_build_environment(self, *, targets: list[str] | None = None, keep_env: bool = False) -> None:
        from hatch.project.constants import BUILD_BACKEND, BuildEnvVars
        from hatch.utils.structures import EnvVars

        if targets is None:
            targets = ["wheel"]

        env_vars = {BuildEnvVars.REQUESTED_TARGETS: " ".join(sorted(targets))}
        build_backend = self.metadata.build.build_backend
        with self.location.as_cwd(), self.build_env.get_env_vars(), EnvVars(env_vars):
            if not self.build_env.exists():
                try:
                    self.build_env.check_compatibility()
                except Exception as e:  # noqa: BLE001
                    self.app.abort(f"Environment `{self.build_env.name}` is incompatible: {e}")

            self.prepare_environment(self.build_env, keep_env=keep_env)

            additional_dependencies: list[str] = []
            with self.app.status("Inspecting build dependencies"):
                if build_backend != BUILD_BACKEND:
                    for target in targets:
                        if target == "sdist":
                            additional_dependencies.extend(self.build_frontend.get_requires("sdist"))
                        elif target == "wheel":
                            additional_dependencies.extend(self.build_frontend.get_requires("wheel"))
                        else:
                            self.app.abort(f"Target `{target}` is not supported by `{build_backend}`")
                else:
                    required_build_deps = self.build_frontend.hatch.get_required_build_deps(targets)
                    if required_build_deps:
                        with self.metadata.context.apply_context(self.build_env.context):
                            additional_dependencies.extend(
                                self.metadata.context.format(dep) for dep in required_build_deps
                            )

            if additional_dependencies:
                from hatch.dep.core import Dependency

                self.build_env.additional_dependencies.extend(map(Dependency, additional_dependencies))
                with self.build_env.app_status_dependency_synchronization():
                    self.build_env.sync_dependencies()

    def get_dependencies(self) -> tuple[list[str], dict[str, list[str]]]:
        dynamic_fields = {"dependencies", "optional-dependencies"}
        if not dynamic_fields.intersection(self.metadata.dynamic):
            dependencies: list[str] = self.metadata.core_raw_metadata.get("dependencies", [])
            features: dict[str, list[str]] = self.metadata.core_raw_metadata.get("optional-dependencies", {})
            return dependencies, features

        from hatch.project.constants import BUILD_BACKEND

        self.prepare_build_environment()
        build_backend = self.metadata.build.build_backend
        with self.location.as_cwd(), self.build_env.get_env_vars():
            if build_backend != BUILD_BACKEND:
                project_metadata = self.build_frontend.get_core_metadata()
            else:
                project_metadata = self.build_frontend.hatch.get_core_metadata()

        dynamic_dependencies: list[str] = project_metadata.get("dependencies", [])
        dynamic_features: dict[str, list[str]] = project_metadata.get("optional-dependencies", {})

        return dynamic_dependencies, dynamic_features

    @cached_property
    def has_static_dependencies(self) -> bool:
        dynamic_fields = {"dependencies", "optional-dependencies"}
        return not dynamic_fields.intersection(self.metadata.dynamic)

    def expand_environments(self, env_name: str) -> list[str]:
        if env_name in self.config.internal_matrices:
            return list(self.config.internal_matrices[env_name]["envs"])

        if env_name in self.config.matrices:
            return list(self.config.matrices[env_name]["envs"])

        if env_name in self.config.internal_envs:
            return [env_name]

        if env_name in self.config.envs:
            return [env_name]

        return []

    @classmethod
    def from_config(cls, config: RootConfig, project: str) -> Project | None:
        # Disallow empty strings
        if not project:
            return None

        if project in config.projects:
            location = config.projects[project].location
            if location:
                return cls(Path(location).resolve(), name=project)
        else:
            for project_dir in config.dirs.project:
                if not project_dir:
                    continue

                location = Path(project_dir, project)
                if location.is_dir():
                    return cls(Path(location).resolve(), name=project)

        return None

    def find_project_root(self) -> Path | None:
        path = self._path

        while True:
            possible_file = path.joinpath("pyproject.toml")
            if possible_file.is_file():
                self._project_file_path = possible_file
                return path

            if path.joinpath("setup.py").is_file():
                return path

            new_path = path.parent
            if new_path == path:
                return None

            path = new_path

    @contextmanager
    def ensure_cwd(self) -> Generator[Path, None, None]:
        cwd = Path.cwd()
        location = self.location
        if location.is_file() or cwd == location or location in cwd.parents:
            yield cwd
        else:
            with location.as_cwd():
                yield location

    @staticmethod
    def canonicalize_name(name: str, *, strict=True) -> str:
        if strict:
            return re.sub(r"[-_.]+", "-", name).lower()

        # Used for creating new projects
        return re.sub(r"[-_. ]+", "-", name).lower()

    @property
    def metadata(self):
        if self._metadata is None:
            from hatchling.metadata.core import ProjectMetadata

            self._metadata = ProjectMetadata(self.location, self.plugin_manager, self.raw_config)

        return self._metadata

    @property
    def raw_config(self):
        if self._raw_config is None:
            if self.root is None or self._project_file_path is None:
                # Assume no pyproject.toml e.g. environment management only
                self._raw_config = {"project": {"name": self.location.name}}
            else:
                from hatch.utils.toml import load_toml_file

                raw_config = load_toml_file(str(self._project_file_path))
                # Assume environment management only
                if "project" not in raw_config:
                    raw_config["project"] = {"name": self.location.name}

                self._raw_config = raw_config

        return self._raw_config

    def save_config(self, config):
        import tomlkit

        with open(str(self._project_file_path), "w", encoding="utf-8") as f:
            f.write(tomlkit.dumps(config))

    @staticmethod
    def initialize(project_file_path, template_config):
        import tomlkit

        with open(str(project_file_path), encoding="utf-8") as f:
            raw_config = tomlkit.parse(f.read())

        build_system_config = raw_config.setdefault("build-system", {})

        build_system_config.clear()
        build_system_config["requires"] = ["hatchling"]
        build_system_config["build-backend"] = "hatchling.build"

        project_config = raw_config.get("project")
        if project_config is None:
            raw_config["project"] = project_config = {}

        project_name = project_config.get("name")
        if not project_name:
            project_config["name"] = template_config["project_name_normalized"]

        project_description = project_config.get("description")
        if not project_description:
            project_config["description"] = template_config["description"]

        project_config["dynamic"] = ["version"]

        tool_config = raw_config.get("tool")
        if tool_config is None:
            raw_config["tool"] = tool_config = {}

        hatch_config = tool_config.get("hatch")
        if hatch_config is None:
            tool_config["hatch"] = hatch_config = {}

        version_config = hatch_config.get("version")
        if version_config is None:
            hatch_config["version"] = version_config = {}

        version_config.clear()
        version_config["path"] = f"{template_config['package_name']}/__init__.py"

        with open(str(project_file_path), "w", encoding="utf-8") as f:
            f.write(tomlkit.dumps(raw_config))



from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

from hatch.env.lockers.interface import LockerInterface
from hatch.env.utils import add_verbosity_flag
from hatch.utils.fs import Path

if TYPE_CHECKING:
    from hatch.env.plugin.interface import EnvironmentInterface


class PipLocker(LockerInterface):
    PLUGIN_NAME = "pip"

    @classmethod
    def supports(cls, _environment: EnvironmentInterface) -> bool:
        return True

    @classmethod
    def generate(
        cls,
        environment: EnvironmentInterface,
        dependencies: list[str],
        output_path: Path,
        *,
        upgrade: bool = False,
        upgrade_packages: tuple[str, ...] = (),
        layered: bool = False,
        lock_extras: tuple[str, ...] = (),
        lock_groups: tuple[str, ...] = (),
    ) -> None:
        if layered or lock_extras or lock_groups:
            message = "The pip locker does not support layered locks with extras or dependency-groups; use installer uv"
            raise ValueError(message)

        if not dependencies:
            return

        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False, encoding="utf-8") as f:
            f.write("\n".join(dependencies))
            f.write("\n")
            deps_file = Path(f.name)

        try:
            command = ["python", "-u", "-m", "pip", "lock", "-r", str(deps_file), "-o", str(output_path)]

            add_verbosity_flag(command, environment.verbosity, adjustment=-1)

            if upgrade:
                command.append("--upgrade")
            for pkg in upgrade_packages:
                command.extend(["--upgrade-package", pkg])

            with environment.command_context():
                environment.platform.check_command(command)
        finally:
            deps_file.unlink(missing_ok=True)

    @classmethod
    def in_sync(
        cls,
        environment: EnvironmentInterface,
        dependencies: list[str],
        output_path: Path,
        *,
        upgrade: bool = False,
        upgrade_packages: tuple[str, ...] = (),
        layered: bool = False,
        lock_extras: tuple[str, ...] = (),
        lock_groups: tuple[str, ...] = (),
    ) -> bool:
        if not output_path.is_file():
            return False

        if layered or lock_extras or lock_groups:
            return False

        with tempfile.NamedTemporaryFile(mode="w", suffix=".toml", delete=False, encoding="utf-8") as f:
            temp_path = Path(f.name)

        try:
            cls.generate(
                environment,
                dependencies,
                temp_path,
                upgrade=upgrade,
                upgrade_packages=upgrade_packages,
                layered=False,
                lock_extras=(),
                lock_groups=(),
            )
            existing = output_path.read_text(encoding="utf-8")
            fresh = temp_path.read_text(encoding="utf-8")
        finally:
            temp_path.unlink(missing_ok=True)

        return existing == fresh

    @classmethod
    def apply_lock(cls, _environment: EnvironmentInterface, _lock_path: Path) -> None:
        from hatch.env.lock import LockerUnsupportedError

        message = (
            "Applying lockfiles with the pip locker is not yet supported. "
            'Use `installer = "uv"` (or `locker = "uv"`) for locked sync.'
        )
        raise LockerUnsupportedError(cls.PLUGIN_NAME, detail=message)

    @classmethod
    def install_matches_lock(cls, _environment: EnvironmentInterface, _lock_path: Path) -> bool:
        # We cannot verify lock conformance for pip because apply_lock is unsupported.
        return False



import click

from hatch.config.constants import PublishEnvVars


@click.command(short_help="Publish build artifacts")
@click.argument("artifacts", nargs=-1)
@click.option(
    "--repo",
    "-r",
    envvar=PublishEnvVars.REPO,
    help="The repository with which to publish artifacts [env var: `HATCH_INDEX_REPO`]",
)
@click.option(
    "--user", "-u", envvar=PublishEnvVars.USER, help="The user with which to authenticate [env var: `HATCH_INDEX_USER`]"
)
@click.option(
    "--auth",
    "-a",
    envvar=PublishEnvVars.AUTH,
    help="The credentials to use for authentication [env var: `HATCH_INDEX_AUTH`]",
)
@click.option(
    "--ca-cert",
    envvar=PublishEnvVars.CA_CERT,
    help="The path to a CA bundle [env var: `HATCH_INDEX_CA_CERT`]",
)
@click.option(
    "--client-cert",
    envvar=PublishEnvVars.CLIENT_CERT,
    help="The path to a client certificate, optionally containing the private key [env var: `HATCH_INDEX_CLIENT_CERT`]",
)
@click.option(
    "--client-key",
    envvar=PublishEnvVars.CLIENT_KEY,
    help="The path to the client certificate's private key [env var: `HATCH_INDEX_CLIENT_KEY`]",
)
@click.option("--no-prompt", "-n", is_flag=True, help="Disable prompts, such as for missing required fields")
@click.option(
    "--initialize-auth", is_flag=True, help="Save first-time authentication information even if nothing was published"
)
@click.option(
    "--publisher",
    "-p",
    "publisher_name",
    envvar=PublishEnvVars.PUBLISHER,
    default="index",
    help="The publisher plugin to use (default is `index`) [env var: `HATCH_PUBLISHER`]",
)
@click.option(
    "--option",
    "-o",
    "options",
    envvar=PublishEnvVars.OPTIONS,
    multiple=True,
    help=(
        "Options to pass to the publisher plugin. This may be selected multiple "
        "times e.g. `-o foo=bar -o baz=23` [env var: `HATCH_PUBLISHER_OPTIONS`]"
    ),
)
@click.option("--yes", "-y", is_flag=True, help="Confirm without prompting when the plugin is disabled")
@click.pass_obj
def publish(
    app,
    artifacts,
    repo,
    user,
    auth,
    ca_cert,
    client_cert,
    client_key,
    no_prompt,
    initialize_auth,
    publisher_name,
    options,
    yes,
):
    """Publish build artifacts."""
    option_map = {"no_prompt": no_prompt, "initialize_auth": initialize_auth}
    if publisher_name == "index":
        if options:
            app.abort("Use the standard CLI flags rather than passing explicit options when using the `index` plugin")

        if repo:
            option_map["repo"] = repo
        if user:
            option_map["user"] = user
        if auth:
            option_map["auth"] = auth
        if ca_cert:
            option_map["ca_cert"] = ca_cert
        if client_cert:
            option_map["client_cert"] = client_cert
        if client_key:
            option_map["client_key"] = client_key
    else:  # no cov
        for option in options:
            key, _, value = option.partition("=")
            option_map[key] = value

    publisher_class = app.plugins.publisher.get(publisher_name)
    if publisher_class is None:
        app.abort(f"Unknown publisher: {publisher_name}")

    publisher = publisher_class(
        app,
        app.project.location,
        app.cache_dir / "publish" / publisher_name,
        app.project.config.publish.get(publisher_name, {}),
        app.config.publish.get(publisher_name, {}),
    )
    if publisher.disable and not (yes or (not no_prompt and app.confirm(f"Confirm `{publisher_name}` publishing"))):
        app.abort(f"Publisher is disabled: {publisher_name}")

    publisher.publish(list(artifacts), option_map)



import os

import click

from hatch.cli.self.report import report

__management_command = os.environ.get("PYAPP_COMMAND_NAME", "self")


@click.group(name=__management_command, short_help="Manage Hatch")
def self_command():
    pass


self_command.add_command(report)

if __management_command:
    from hatch.cli.self.restore import restore
    from hatch.cli.self.update import update

    self_command.add_command(restore)
    self_command.add_command(update)



from __future__ import annotations

import os
import pathlib
import sys
from contextlib import contextmanager, suppress
from functools import cached_property
from typing import TYPE_CHECKING, Any

from hatch.utils.structures import EnvVars

if TYPE_CHECKING:
    from collections.abc import Generator

    from _typeshed import FileDescriptorLike


# There is special recognition in Mypy for `sys.platform`, not `os.name`
# https://github.com/python/cpython/blob/09d7319bfe0006d9aa3fc14833b69c24ccafdca6/Lib/pathlib.py#L957
if sys.platform == "win32":
    _PathBase = pathlib.WindowsPath
else:
    _PathBase = pathlib.PosixPath

disk_sync = os.fsync
# https://mjtsai.com/blog/2022/02/17/apple-ssd-benchmarks-and-f_fullsync/
# https://developer.apple.com/library/archive/documentation/System/Conceptual/ManPages_iPhoneOS/man2/fsync.2.html
if sys.platform == "darwin":
    import fcntl

    if hasattr(fcntl, "F_FULLFSYNC"):

        def disk_sync(fd: FileDescriptorLike) -> None:
            fcntl.fcntl(fd, fcntl.F_FULLFSYNC)


class Path(_PathBase):
    @cached_property
    def long_id(self) -> str:
        from base64 import urlsafe_b64encode
        from hashlib import sha256

        path = str(self)
        if sys.platform == "win32" or sys.platform == "darwin":
            path = path.casefold()

        digest = sha256(path.encode("utf-8")).digest()
        return urlsafe_b64encode(digest).decode("utf-8")

    @cached_property
    def id(self) -> str:
        return self.long_id[:8]

    def ensure_dir_exists(self) -> None:
        self.mkdir(parents=True, exist_ok=True)

    def ensure_parent_dir_exists(self) -> None:
        self.parent.mkdir(parents=True, exist_ok=True)

    def expand(self) -> Path:
        return Path(os.path.expanduser(os.path.expandvars(self)))

    def remove(self) -> None:
        if self.is_file():
            os.remove(self)
        elif self.is_dir():
            import shutil

            shutil.rmtree(self, ignore_errors=False)

    def move(self, target):
        try:
            self.replace(target)
        except OSError:
            import shutil

            shutil.copy2(self, target)
            self.unlink()

    def wait_for_dir_removed(self, timeout: int = 5) -> None:
        import shutil
        import time

        for _ in range(timeout * 2):
            if self.is_dir():
                shutil.rmtree(self, ignore_errors=True)
                time.sleep(0.5)
            else:
                return

        if self.is_dir():
            shutil.rmtree(self, ignore_errors=False)

    def write_atomic(self, data: str | bytes, *args: Any, **kwargs: Any) -> None:
        from tempfile import mkstemp

        fd, path = mkstemp(dir=self.parent)
        with os.fdopen(fd, *args, **kwargs) as f:
            f.write(data)
            f.flush()
            disk_sync(fd)

        os.replace(path, self)

    @contextmanager
    def as_cwd(self, *args: Any, **kwargs: Any) -> Generator[Path, None, None]:
        origin = os.getcwd()
        os.chdir(self)

        try:
            if args or kwargs:
                with EnvVars(*args, **kwargs):
                    yield self
            else:
                yield self
        finally:
            os.chdir(origin)

    @contextmanager
    def temp_hide(self) -> Generator[Path, None, None]:
        import shutil

        with temp_directory() as temp_dir:
            temp_path = Path(temp_dir, self.name)
            with suppress(FileNotFoundError):
                shutil.move(str(self), temp_dir / self.name)

            try:
                yield temp_path
            finally:
                with suppress(FileNotFoundError):
                    shutil.move(str(temp_path), self)

    if sys.platform == "win32":

        @classmethod
        def from_uri(cls, path: str) -> Path:
            return cls(path.replace("file:///", "", 1))

    else:

        @classmethod
        def from_uri(cls, path: str) -> Path:
            return cls(path.replace("file://", "", 1))


@contextmanager
def temp_directory() -> Generator[Path, None, None]:
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as d:
        yield Path(d).resolve()


@contextmanager
def temp_chdir(env_vars: dict[str, str] | None = None) -> Generator[Path, None, None]:
    with temp_directory() as d, d.as_cwd(env_vars=env_vars):
        yield d



from __future__ import annotations

from typing import Any


def get_default_config() -> dict[str, Any]:
    return {
        "skip-install": True,
        "builder": True,
        "installer": "uv",
    }



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(
    short_help="Run commands within project environments",
    context_settings={"help_option_names": [], "ignore_unknown_options": True},
)
@click.argument("args", metavar="[ENV:]ARGS...", required=True, nargs=-1)
@click.pass_context
def run(ctx: click.Context, args: tuple[str, ...]):
    """
    Run commands within project environments.
    This is a convenience wrapper around the [`env run`](#hatch-env-run) command.

    If the first argument contains a colon, then the preceding component will be
    interpreted as the name of the environment to target, overriding the `-e`/`--env`
    [root option](#hatch) and the `HATCH_ENV` environment variable.

    If the environment provides matrices, then you may also provide leading arguments
    starting with a `+` or `-` to select or exclude certain variables, optionally
    followed by specific comma-separated values. For example, if you have the
    following configuration:

    \b
    ```toml tab="pyproject.toml"
    [[tool.hatch.envs.test.matrix]]
    python = ["3.9", "3.10"]
    version = ["42", "3.14", "9000"]
    ```

    \b
    ```toml tab="hatch.toml"
    [[envs.test.matrix]]
    python = ["3.9", "3.10"]
    version = ["42", "3.14", "9000"]
    ```

    then running:

    \b
    ```
    hatch run +py=3.10 -version=9000 test:pytest
    ```

    would execute `pytest` in the environments `test.py3.10-42` and `test.py3.10-3.14`.
    Note that `py` may be used as an alias for `python`.

    \b
    !!! note
        Inclusions are treated as an intersection while exclusions are treated as a union i.e.
        an environment must match all of the included variables to be selected while matching
        any of the excluded variables will prevent selection.
    """
    app: Application = ctx.obj

    first_arg = args[0]
    if first_arg in {"-h", "--help"}:
        app.display_info(ctx.get_help())
        return

    from hatch.utils.fs import Path

    if first_arg.endswith(".py") and (script := Path(first_arg)).is_file():
        from hatch.project.utils import parse_inline_script_metadata

        # Ensure consistent IDs for storage
        script = script.resolve()

        try:
            metadata = parse_inline_script_metadata(script.read_text())
        except ValueError as e:
            app.abort(f"{e}, {first_arg}")

        # Ignore scripts that don't define metadata blocks or define empty metadata blocks
        if metadata:
            from hatch.env.utils import ensure_valid_environment

            config = metadata.get("tool", {}).get("hatch", {})
            config["skip-install"] = True
            config.setdefault("installer", "uv")
            config.setdefault("dependencies", [])
            config["dependencies"].extend(metadata.get("dependencies", []))

            if "python" not in config and (requires_python := metadata.get("requires-python")) is not None:
                import re
                import sys
                import sysconfig

                from packaging.specifiers import SpecifierSet

                from hatch.python.distributions import DISTRIBUTIONS

                current_version = ".".join(map(str, sys.version_info[:2]))
                if bool(sysconfig.get_config_var("Py_GIL_DISABLED")):
                    current_version += "t"

                # Strip "t" suffix for distribution lookup since DISTRIBUTIONS keys don't include it
                current_version_base = current_version.rstrip("t")
                distributions = [name for name in DISTRIBUTIONS if re.match(r"^\d+\.\d+$", name)]
                distributions.sort(key=lambda name: name != current_version_base)

                python_constraint = SpecifierSet(requires_python)
                for distribution in distributions:
                    # Try an artificially high patch version to account for
                    # common cases like `>=3.11.4` or `>=3.10,<3.11`
                    if python_constraint.contains(f"{distribution}.100"):
                        # Only set config["python"] if it doesn't match the current Python's base version
                        # This allows free-threaded builds (e.g. 3.14t) to match their base version (3.14)
                        if distribution != current_version_base:
                            config["python"] = distribution
                        break
                else:
                    app.abort(f"Unable to satisfy Python version constraint: {requires_python}")

            ensure_valid_environment(config)
            app.project.config.envs[script.id] = config
            app.project.set_path(script)
            for context in app.runner_context([script.id]):
                context.add_shell_command(["python", first_arg, *args[1:]])

            return

    from hatch.cli.env.run import run as run_command

    command_start = 0
    included_variables = []
    excluded_variables = []
    for i, arg in enumerate(args):
        command_start = i
        if arg.startswith("+"):
            included_variables.append(arg[1:])
        elif arg.startswith("-"):
            excluded_variables.append(arg[1:])
        else:
            break
    else:
        command_start += 1

    args = args[command_start:]
    if not args:
        app.abort("Missing argument `MATRIX:ARGS...`")

    command, *final_args = args
    env_name, separator, command = command.rpartition(":")
    if not separator:
        env_name = app.env
    elif not env_name:
        env_name = "system"

    ctx.invoke(
        run_command,
        args=[command, *final_args],
        env_names=[env_name],
        included_variable_specs=included_variables,
        excluded_variable_specs=excluded_variables,
    )



from __future__ import annotations

import time
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Generator

    import httpx2

    from hatch.utils.fs import Path

MINIMUM_SLEEP = 2
MAXIMUM_SLEEP = 20
# The timeout should be slightly larger than a multiple of 3,
# which is the default TCP packet retransmission window. See:
# https://tools.ietf.org/html/rfc2988
DEFAULT_TIMEOUT = 10


@contextmanager
def streaming_response(*args: Any, **kwargs: Any) -> Generator[httpx2.Response, None, None]:
    from secrets import choice

    import httpx2

    attempts = 0
    while True:
        attempts += 1
        try:
            with httpx2.stream(*args, **kwargs) as response:
                response.raise_for_status()
                yield response

            break
        except httpx2.HTTPError:
            sleep = min(MAXIMUM_SLEEP, MINIMUM_SLEEP * 2**attempts)
            if sleep == MAXIMUM_SLEEP:
                raise

            time.sleep(choice(range(sleep + 1)))


def download_file(path: Path, *args: Any, **kwargs: Any) -> None:
    kwargs.setdefault("timeout", DEFAULT_TIMEOUT)

    with path.open(mode="wb", buffering=0) as f, streaming_response("GET", *args, **kwargs) as response:
        for chunk in response.iter_bytes(16384):
            f.write(chunk)



import os
import sys

FILE = os.path.abspath(__file__)
HERE = os.path.dirname(FILE)
ENV_VAR_PREFIX = "_HATCHLING_PORT_ADD_"


def _apply_env_vars(kwargs):
    from ast import literal_eval

    for key, value in os.environ.items():
        if key.startswith(ENV_VAR_PREFIX):
            kwargs[key.replace(ENV_VAR_PREFIX, "", 1).lower()] = literal_eval(value)


def _parse_dependencies(dependency_definition):
    dependencies = []
    for line in dependency_definition.splitlines():
        dependency = line.split(" #", 1)[0].strip()
        if dependency:
            dependencies.append(dependency)

    return dependencies


def _collapse_data(output, data):
    import tomli_w

    expected_output = new_output = tomli_w.dumps(data)
    new_output = new_output.replace("    ", "")
    new_output = new_output.replace("[\n", "[")
    new_output = new_output.replace('",\n]', '"]')

    return output.replace(expected_output, new_output, 1)


def _parse_setup_cfg(kwargs):
    from configparser import ConfigParser

    setup_cfg_file = os.path.join(HERE, "setup.cfg")
    if not os.path.isfile(setup_cfg_file):
        return

    setup_cfg = ConfigParser()
    setup_cfg.read(setup_cfg_file, encoding="utf-8")

    if not setup_cfg.has_section("metadata"):
        return

    metadata = setup_cfg["metadata"]

    if "name" in metadata and "name" not in kwargs:
        kwargs["name"] = metadata["name"]

    if "description" in metadata and "description" not in kwargs:
        kwargs["description"] = metadata["description"]

    if "license" in metadata and "license" not in kwargs:
        kwargs["license"] = metadata["license"]

    if "author" in metadata and "author" not in kwargs:
        kwargs["author"] = metadata["author"]

    if "author_email" in metadata and "author_email" not in kwargs:
        kwargs["author_email"] = metadata["author_email"]

    if "keywords" in metadata and "keywords" not in kwargs:
        keywords = metadata["keywords"].strip().splitlines()
        kwargs["keywords"] = keywords if len(keywords) > 1 else keywords[0]

    if "classifiers" in metadata and "classifiers" not in kwargs:
        kwargs["classifiers"] = metadata["classifiers"].strip().splitlines()

    if "url" in metadata and "url" not in kwargs:
        kwargs["url"] = metadata["url"]

    if "download_url" in metadata and "download_url" not in kwargs:
        kwargs["download_url"] = metadata["download_url"]

    if "project_urls" in metadata and "project_urls" not in kwargs:
        kwargs["project_urls"] = dict(
            url.replace(" = ", "=", 1).split("=") for url in metadata["project_urls"].strip().splitlines()
        )

    if setup_cfg.has_section("options"):
        options = setup_cfg["options"]

        if "python_requires" in options and "python_requires" not in kwargs:
            kwargs["python_requires"] = options["python_requires"]

        if "install_requires" in options and "install_requires" not in kwargs:
            kwargs["install_requires"] = _parse_dependencies(options["install_requires"])

        if "packages" in options and "packages" not in kwargs:
            packages = []
            for package_spec in options["packages"].strip().splitlines():
                package = package_spec.replace("find:", "", 1).replace("find_namespace:", "", 1).strip()
                if package:
                    packages.append(package)

            if packages:
                kwargs["packages"] = packages

        if "package_dir" in options and "package_dir" not in kwargs:
            kwargs["package_dir"] = dict(
                d.replace(" = ", "=", 1).split("=") for d in options["package_dir"].strip().splitlines()
            )

    if setup_cfg.has_section("options.extras_require") and "extras_require" not in kwargs:
        kwargs["extras_require"] = {
            feature: _parse_dependencies(dependencies)
            for feature, dependencies in setup_cfg["options.extras_require"].items()
        }

    if setup_cfg.has_section("options.entry_points") and "entry_points" not in kwargs:
        kwargs["entry_points"] = {
            entry_point: definitions.strip().splitlines()
            for entry_point, definitions in setup_cfg["options.entry_points"].items()
        }


def setup(**kwargs):
    import itertools
    import re

    import tomli_w

    project_metadata = {}
    hatch_metadata = {}
    tool_metadata = {"hatch": hatch_metadata}
    project_data = {
        "build-system": {"requires": ["hatchling"], "build-backend": "hatchling.build"},
        "project": project_metadata,
        "tool": tool_metadata,
    }

    _parse_setup_cfg(kwargs)
    _apply_env_vars(kwargs)

    name = kwargs["name"]
    project_name = name.replace("_", "-")
    packages = sorted(kwargs.get("packages") or [name.replace("-", "_")])
    package_name = package_path = package_source = packages[0].split(".")[0].lower().strip()

    project_metadata["name"] = project_name

    project_metadata["dynamic"] = ["version"]

    if "description" in kwargs:
        project_metadata["description"] = kwargs["description"]

    for readme_file in ("README.md", "README.rst", "README.txt"):
        if os.path.isfile(os.path.join(HERE, readme_file)):
            project_metadata["readme"] = readme_file
            break

    project_metadata["license"] = kwargs.get("license", "")

    if "python_requires" in kwargs:
        project_metadata["requires-python"] = kwargs["python_requires"]

    for collaborator in ("author", "maintainer"):
        collaborators = []
        collaborator_names = []
        collaborator_emails = []
        if collaborator in kwargs:
            collaborator_names.extend(
                collaborator_name.strip() for collaborator_name in kwargs[collaborator].split(",")
            )
        if f"{collaborator}_email" in kwargs:
            collaborator_emails.extend(
                collaborator_email.strip() for collaborator_email in kwargs[f"{collaborator}_email"].split(",")
            )

        for collaborator_name, collaborator_email in itertools.zip_longest(collaborator_names, collaborator_emails):
            data = {}
            if collaborator_name is not None:
                data["name"] = collaborator_name
            if collaborator_email is not None:
                data["email"] = collaborator_email
            if data:
                collaborators.append(data)

        if collaborators:
            project_metadata[f"{collaborator}s"] = collaborators

    if "keywords" in kwargs:
        keywords = kwargs["keywords"]
        if isinstance(keywords, str):
            keywords = keywords.replace(",", " ").split()
        project_metadata["keywords"] = sorted(keywords)

    if "classifiers" in kwargs:
        project_metadata["classifiers"] = sorted(kwargs["classifiers"])
        fixed_indices = []
        final_index = 0
        for i, classifier in enumerate(project_metadata["classifiers"]):
            if classifier.startswith("Programming Language :: Python :: "):
                final_index = i
                for python_version in ("3.10", "3.11", "3.12"):
                    if classifier.endswith(python_version):
                        fixed_indices.append(i)
                        break
        for i, index in enumerate(fixed_indices):
            project_metadata["classifiers"].insert(final_index, project_metadata["classifiers"].pop(index - i))

    if "install_requires" in kwargs:
        project_metadata["dependencies"] = sorted(
            [entry.strip() for entry in kwargs["install_requires"]]
            if isinstance(kwargs["install_requires"], (list, tuple))
            else _parse_dependencies(kwargs["install_requires"]),
            key=lambda d: d.lower(),
        )

    if "extras_require" in kwargs:
        project_metadata["optional-dependencies"] = {
            group: sorted(
                [entry.strip() for entry in dependencies]
                if isinstance(dependencies, (list, tuple))
                else _parse_dependencies(dependencies),
                key=lambda d: d.lower(),
            )
            for group, dependencies in sorted(kwargs["extras_require"].items())
        }

    if "entry_points" in kwargs and isinstance(kwargs["entry_points"], dict):
        entry_points = {}
        for entry_point, raw_definitions in kwargs["entry_points"].items():
            definitions = [raw_definitions] if isinstance(raw_definitions, str) else raw_definitions
            definitions = dict(sorted(d.replace(" ", "").split("=", 1) for d in definitions))

            if entry_point == "console_scripts":
                project_metadata["scripts"] = definitions
            elif entry_point == "gui_scripts":
                project_metadata["gui-scripts"] = definitions
            else:
                entry_points[entry_point] = definitions
        if entry_points:
            project_metadata["entry-points"] = dict(sorted(entry_points.items()))

    urls = {}
    if "url" in kwargs:
        urls["Homepage"] = kwargs["url"]
    if "download_url" in kwargs:
        urls["Download"] = kwargs["download_url"]
    if "project_urls" in kwargs:
        urls.update(kwargs["project_urls"])
    if urls:
        project_metadata["urls"] = dict(sorted(urls.items()))

    build_targets = {}
    build_data = {}

    if "use_scm_version" in kwargs:
        project_data["build-system"]["requires"].append("hatch-vcs")
        hatch_metadata["version"] = {"source": "vcs"}
        build_data["hooks"] = {"vcs": {"version-file": f"{package_path}/_version.py"}}
    else:
        hatch_metadata["version"] = {"path": f"{package_path}/__init__.py"}

    build_data["targets"] = build_targets

    if "" in kwargs.get("package_dir", {}):
        package_source = kwargs["package_dir"][""]

        package = (kwargs.get("packages") or [package_name])[0]
        package_path = f"{package_source}/{package}"

        if package_path != f"src/{package_name}":
            build_targets.setdefault("wheel", {})["packages"] = [package_path]

    if kwargs.get("data_files", []):
        shared_data = {}
        for shared_directory, relative_paths in kwargs["data_files"]:
            relative_files = {}
            for relative_path in relative_paths:
                relative_directory, filename = os.path.split(relative_path)
                relative_files.setdefault(relative_directory, []).append(filename)

            for relative_directory, files in sorted(relative_files.items()):
                if not os.path.isdir(relative_directory) or set(os.listdir(relative_directory)) != set(files):
                    for filename in sorted(files):
                        local_path = os.path.join(relative_directory, filename).replace("\\", "/")
                        shared_data[local_path] = f"{shared_directory}/{filename}"
                else:
                    shared_data[relative_directory] = shared_directory

        build_targets.setdefault("wheel", {})["shared-data"] = shared_data

    build_targets["sdist"] = {
        "include": [
            f"/{package_source}",
        ]
    }

    hatch_metadata["build"] = build_data

    output = tomli_w.dumps(project_data)
    output = _collapse_data(output, {"requires": project_data["build-system"]["requires"]})
    output = _collapse_data(output, {"dynamic": project_data["project"]["dynamic"]})

    project_file = os.path.join(HERE, "pyproject.toml")
    if os.path.isfile(project_file):
        with open(project_file, encoding="utf-8") as f:
            current_contents = f.read()

        for section in ("build-system", "project"):
            for pattern in (rf"^\[{section}].*?(?=^\[)", rf"^\[{section}].*"):
                current_contents = re.sub(pattern, "", current_contents, flags=re.MULTILINE | re.DOTALL)

        output += f"\n{current_contents}"

    with open(project_file, "w", encoding="utf-8") as f:
        f.write(output)


if __name__ == "setuptools":
    __this_shim = sys.modules.pop("setuptools")
    __current_directory = sys.path.pop(0)

    import setuptools as __real_setuptools

    sys.path.insert(0, __current_directory)
    sys.modules["setuptools"] = __this_shim

    def __getattr__(name):
        return getattr(__real_setuptools, name)

    del __this_shim
    del __current_directory


def migrate(root, setuptools_options, sys_paths):
    import shutil
    import subprocess
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as temp_dir:
        repo_dir = os.path.join(os.path.realpath(temp_dir), "repo")
        shutil.copytree(root, repo_dir, ignore=shutil.ignore_patterns(".git", ".tox"), copy_function=shutil.copy)
        shutil.copy(FILE, os.path.join(repo_dir, "setuptools.py"))
        setup_py = os.path.join(repo_dir, "setup.py")

        if not os.path.isfile(setup_py):
            # Synthesize a small setup.py file since there is none
            with open(setup_py, "w", encoding="utf-8") as f:
                f.write("from setuptools import setup\nsetup()\n")

        env = dict(os.environ)
        for arg in setuptools_options:
            key, value = arg.split("=", 1)
            env[f"{ENV_VAR_PREFIX}{key}"] = value

        # When PYTHONSAFEPATH is non-empty, the current directory is not added automatically
        python_paths = [repo_dir]
        python_paths.extend(p for p in sys_paths if p)
        if python_path := env.get("PYTHONPATH", ""):
            python_paths.append(python_path)

        env["PYTHONPATH"] = os.pathsep.join(python_paths)

        subprocess.run([sys.executable, setup_py], env=env, cwd=repo_dir, check=True)

        old_project_file = os.path.join(root, "pyproject.toml")
        new_project_file = os.path.join(repo_dir, "pyproject.toml")
        shutil.copyfile(new_project_file, old_project_file)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

from hatch.cli.python.install import install

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Update Python distributions")
@click.argument("names", required=True, nargs=-1)
@click.option("--dir", "-d", "directory", help="The directory in which distributions reside")
@click.pass_context
def update(ctx: click.Context, *, names: tuple[str, ...], directory: str | None):
    """
    Update Python distributions.

    You may select `all` to update all installed distributions:

    \b
    ```
    hatch python update all
    ```
    """
    app: Application = ctx.obj

    manager = app.get_python_manager(directory)
    installed = manager.get_installed()
    selection = tuple(installed) if "all" in names else names

    not_installed = [name for name in selection if name not in installed]
    if not_installed:
        app.abort(f"Distributions not installed: {', '.join(not_installed)}")

    ctx.invoke(install, names=selection, directory=directory, private=True, update=True)



from hatchling.plugin.specs import hookspec


@hookspec
def hatch_register_environment():
    """Register new classes that adhere to the environment interface."""


@hookspec
def hatch_register_environment_collector():
    """Register new classes that adhere to the environment collector interface."""


@hookspec
def hatch_register_version_scheme():
    """Register new classes that adhere to the version scheme interface."""


@hookspec
def hatch_register_publisher():
    """Register new classes that adhere to the publisher interface."""


@hookspec
def hatch_register_template():
    """Register new classes that adhere to the template interface."""


@hookspec
def hatch_register_locker():
    """Register new classes that adhere to the dependency locker interface."""



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.utils.fs import Path


class LockerInterface(ABC):
    """
    Pluggable dependency locker.

    Implementations are registered via the ``hatch_register_locker`` hook.
    """

    PLUGIN_NAME: str = ""

    @classmethod
    @abstractmethod
    def supports(cls, environment: EnvironmentInterface) -> bool:
        """Whether this locker can operate in the given environment."""

    @classmethod
    @abstractmethod
    def generate(
        cls,
        environment: EnvironmentInterface,
        dependencies: list[str],
        output_path: Path,
        *,
        upgrade: bool = False,
        upgrade_packages: tuple[str, ...] = (),
        layered: bool = False,
        lock_extras: tuple[str, ...] = (),
        lock_groups: tuple[str, ...] = (),
    ) -> None:
        """
        Resolve and write a lockfile from PEP 508 ``dependencies`` and optional layered inputs.

        ``dependencies`` is the list of requirement lines (merged env / project inputs) that the
        orchestrator computed for this lock operation.
        """

    @classmethod
    @abstractmethod
    def in_sync(
        cls,
        environment: EnvironmentInterface,
        dependencies: list[str],
        output_path: Path,
        *,
        upgrade: bool = False,
        upgrade_packages: tuple[str, ...] = (),
        layered: bool = False,
        lock_extras: tuple[str, ...] = (),
        lock_groups: tuple[str, ...] = (),
    ) -> bool:
        """
        Return whether ``output_path`` matches what :meth:`generate` would write now.
        """

    @classmethod
    @abstractmethod
    def apply_lock(cls, environment: EnvironmentInterface, lock_path: Path) -> None:
        """Install packages so the environment matches the lockfile at ``lock_path``."""

    @classmethod
    def install_matches_lock(cls, _environment: EnvironmentInterface, _lock_path: Path) -> bool:
        """
        Whether the environment already matches what :meth:`apply_lock` would install.

        Used for ``locked`` environments in :meth:`~hatch.env.virtual.VirtualEnvironment.dependencies_in_sync`.
        """
        return True



from __future__ import annotations

import os
from fnmatch import fnmatch
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import TracebackType


class EnvVars(dict):
    def __init__(
        self, env_vars: dict | None = None, include: list[str] | None = None, exclude: list[str] | None = None
    ) -> None:
        super().__init__(os.environ)
        self.old_env = dict(self)

        if include:
            self.clear()
            for env_var, value in self.old_env.items():
                for pattern in include:
                    if fnmatch(env_var, pattern):
                        self[env_var] = value
                        break

        if exclude:
            for env_var in list(self):
                for pattern in exclude:
                    if fnmatch(env_var, pattern):
                        self.pop(env_var)
                        break

        if env_vars:
            self.update(env_vars)

    def __enter__(self) -> None:
        os.environ.clear()
        os.environ.update(self)

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        os.environ.clear()
        os.environ.update(self.old_env)



from __future__ import annotations

from typing import Any


def get_default_config() -> dict[str, Any]:
    return {
        "skip-install": True,
        "installer": "uv",
        "dependencies": [f"ruff=={RUFF_DEFAULT_VERSION}"],
        "scripts": {
            "format-check": "ruff format{env:HATCH_FMT_ARGS:}{env:HATCH_CHECK_FMT_ARGS:} --check --diff {args:.}",
            "format-fix": "ruff format{env:HATCH_FMT_ARGS:}{env:HATCH_CHECK_FMT_ARGS:} {args:.}",
            "lint-check": "ruff check{env:HATCH_FMT_ARGS:}{env:HATCH_CHECK_CODE_ARGS:} {args:.}",
            "lint-fix": "ruff check{env:HATCH_FMT_ARGS:}{env:HATCH_CHECK_CODE_ARGS:} --fix {args:.}",
        },
    }


def get_check_code_config() -> dict[str, Any]:
    return {
        "skip-install": True,
        "installer": "uv",
        "dependencies": [f"ruff=={RUFF_DEFAULT_VERSION}"],
        "scripts": {
            "lint-check": "ruff check{env:HATCH_CHECK_CODE_ARGS:} {args:.}",
            "lint-fix": "ruff check{env:HATCH_CHECK_CODE_ARGS:} --fix {args:.}",
        },
    }


def get_check_fmt_config() -> dict[str, Any]:
    return {
        "skip-install": True,
        "installer": "uv",
        "dependencies": [f"ruff=={RUFF_DEFAULT_VERSION}"],
        "scripts": {
            "format-check": "ruff format{env:HATCH_CHECK_FMT_ARGS:} --check --diff {args:.}",
            "format-fix": "ruff format{env:HATCH_CHECK_FMT_ARGS:} {args:.}",
        },
    }


RUFF_DEFAULT_VERSION: str = "0.13.2"



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Build a project")
@click.argument("location", required=False)
@click.option(
    "--target",
    "-t",
    "targets",
    multiple=True,
    help=(
        "The target to build, overriding project defaults. This may be selected multiple times e.g. `-t sdist -t wheel`"
    ),
)
@click.option(
    "--hooks-only", is_flag=True, help="Whether or not to only execute build hooks [env var: `HATCH_BUILD_HOOKS_ONLY`]"
)
@click.option(
    "--no-hooks", is_flag=True, help="Whether or not to disable build hooks [env var: `HATCH_BUILD_NO_HOOKS`]"
)
@click.option(
    "--ext",
    is_flag=True,
    help=(
        "Whether or not to only execute build hooks for distributing binary Python packages, such as "
        "compiling extensions. Equivalent to `--hooks-only -t wheel`"
    ),
)
@click.option(
    "--clean",
    "-c",
    is_flag=True,
    help="Whether or not existing artifacts should first be removed [env var: `HATCH_BUILD_CLEAN`]",
)
@click.option(
    "--clean-hooks-after",
    is_flag=True,
    help=(
        "Whether or not build hook artifacts should be removed after each build "
        "[env var: `HATCH_BUILD_CLEAN_HOOKS_AFTER`]"
    ),
)
@click.option("--clean-only", is_flag=True, hidden=True)
@click.pass_obj
def build(app: Application, location, targets, hooks_only, no_hooks, ext, clean, clean_hooks_after, clean_only):
    """Build a project."""
    app.ensure_environment_plugin_dependencies()

    from hatch.config.constants import AppEnvVars
    from hatch.project.config import env_var_enabled
    from hatch.project.constants import BUILD_BACKEND, DEFAULT_BUILD_DIRECTORY, BuildEnvVars
    from hatch.utils.fs import Path
    from hatch.utils.runner import ExecutionContext
    from hatch.utils.structures import EnvVars

    build_dir = Path(location).resolve() if location else None
    if ext:
        hooks_only = True
        targets = ("wheel",)
    elif not targets:
        targets = ("sdist", "wheel")

    env_vars = {}
    if app.verbose:
        env_vars[AppEnvVars.VERBOSE] = str(app.verbosity)
    elif app.quiet:
        env_vars[AppEnvVars.QUIET] = str(abs(app.verbosity))

    with EnvVars(env_vars):
        app.project.prepare_build_environment(targets=[target.split(":")[0] for target in targets])

    build_backend = app.project.metadata.build.build_backend
    with app.project.location.as_cwd(), app.project.build_env.get_env_vars():
        for target in targets:
            target_name, _, _ = target.partition(":")
            if not clean_only:
                app.display_header(target_name)

            if build_backend != BUILD_BACKEND:
                if target_name == "sdist":
                    directory = build_dir or app.project.location / DEFAULT_BUILD_DIRECTORY
                    directory.ensure_dir_exists()
                    artifact_path = app.project.build_frontend.build_sdist(directory)
                elif target_name == "wheel":
                    directory = build_dir or app.project.location / DEFAULT_BUILD_DIRECTORY
                    directory.ensure_dir_exists()
                    artifact_path = app.project.build_frontend.build_wheel(directory)
                else:
                    app.abort(f"Target `{target_name}` is not supported by `{build_backend}`")

                app.display_info(
                    str(artifact_path.relative_to(app.project.location))
                    if app.project.location in artifact_path.parents
                    else str(artifact_path)
                )
            else:
                command = ["python", "-u", "-m", "hatchling", "build", "--target", target]

                # We deliberately pass the location unchanged so that absolute paths may be non-local
                # and reflect wherever builds actually take place
                if location:
                    command.extend(("--directory", location))

                if hooks_only or env_var_enabled(BuildEnvVars.HOOKS_ONLY):
                    command.append("--hooks-only")

                if no_hooks or env_var_enabled(BuildEnvVars.NO_HOOKS):
                    command.append("--no-hooks")

                if clean or env_var_enabled(BuildEnvVars.CLEAN):
                    command.append("--clean")

                if clean_hooks_after or env_var_enabled(BuildEnvVars.CLEAN_HOOKS_AFTER):
                    command.append("--clean-hooks-after")

                if clean_only:
                    command.append("--clean-only")

                context = ExecutionContext(app.project.build_env)
                context.add_shell_command(command)
                context.env_vars.update(env_vars)
                app.execute_context(context)



from __future__ import annotations

from typing import TYPE_CHECKING

import tomlkit

if TYPE_CHECKING:
    from tomlkit.items import InlineTable
    from tomlkit.toml_document import TOMLDocument

    from hatch.utils.fs import Path


def save_toml_document(document: TOMLDocument, path: Path):
    path.ensure_parent_dir_exists()
    path.write_atomic(tomlkit.dumps(document), "w", encoding="utf-8")


def create_toml_document(config: dict) -> InlineTable:
    return tomlkit.item(config)



from __future__ import annotations

import re
import sys
from importlib.metadata import Distribution, DistributionFinder

from packaging.markers import default_environment

from hatch.dep.core import Dependency
from hatch.utils.fs import Path


class InstalledDistributions:
    def __init__(self, *, sys_path: list[str] | None = None, environment: dict[str, str] | None = None) -> None:
        self.__sys_path: list[str] = sys.path if sys_path is None else sys_path
        self.__environment: dict[str, str] = (
            default_environment() if environment is None else environment  # type: ignore[assignment]
        )
        self.__resolver = Distribution.discover(context=DistributionFinder.Context(path=self.__sys_path))
        self.__distributions: dict[str, Distribution] = {}
        self.__search_exhausted = False
        self.__canonical_regex = re.compile(r"[-_.]+")

    def dependencies_in_sync(self, dependencies: list[Dependency]) -> bool:
        return all(self.dependency_in_sync(dependency) for dependency in dependencies)

    def missing_dependencies(self, dependencies: list[Dependency]) -> list[Dependency]:
        return [dependency for dependency in dependencies if not self.dependency_in_sync(dependency)]

    def dependency_in_sync(self, dependency: Dependency, *, environment: dict[str, str] | None = None) -> bool:
        if environment is None:
            environment = self.__environment

        if dependency.marker and not dependency.marker.evaluate(environment):
            return True

        distribution = self[dependency.name]
        if distribution is None:
            return False

        extras = dependency.extras
        if extras:
            transitive_dependencies: list[str] = distribution.metadata.get_all("Requires-Dist", [])
            if not transitive_dependencies:
                return False

            available_extras: list[str] = distribution.metadata.get_all("Provides-Extra", [])

            for dependency_string in transitive_dependencies:
                transitive_dependency = Dependency(dependency_string)
                if not transitive_dependency.marker:
                    continue

                for extra in extras:
                    # FIXME: This may cause a build to never be ready if newer versions do not provide the desired
                    # extra and it's just a user error/typo. See: https://github.com/pypa/pip/issues/7122
                    if extra not in available_extras:
                        return False

                    extra_environment = dict(environment)
                    extra_environment["extra"] = extra
                    if not self.dependency_in_sync(transitive_dependency, environment=extra_environment):
                        return False

        if dependency.specifier and not dependency.specifier.contains(distribution.version):
            return False

        # TODO: handle https://discuss.python.org/t/11938
        if dependency.url:
            direct_url_file = distribution.read_text("direct_url.json")
            if direct_url_file is None:
                return False

            import json

            # https://packaging.python.org/specifications/direct-url/
            direct_url_data = json.loads(direct_url_file)
            url = direct_url_data["url"]
            if "dir_info" in direct_url_data:
                dir_info = direct_url_data["dir_info"]
                editable = dir_info.get("editable", False)
                if editable != dependency.editable:
                    return False

                if Path.from_uri(url) != dependency.path:
                    return False

            if "vcs_info" in direct_url_data:
                vcs_info = direct_url_data["vcs_info"]
                vcs = vcs_info["vcs"]
                commit_id = vcs_info["commit_id"]
                requested_revision = vcs_info.get("requested_revision")

                # Try a few variations, see https://peps.python.org/pep-0440/#direct-references
                if (
                    requested_revision and dependency.url == f"{vcs}+{url}@{requested_revision}#{commit_id}"
                ) or dependency.url == f"{vcs}+{url}@{commit_id}":
                    return True

                if dependency.url in {f"{vcs}+{url}", f"{vcs}+{url}@{requested_revision}"}:
                    import subprocess

                    if vcs == "git":
                        vcs_cmd = [vcs, "ls-remote", url]
                        if requested_revision:
                            vcs_cmd.append(requested_revision)
                    # TODO: add elifs for hg, svn, and bzr https://github.com/pypa/hatch/issues/760
                    else:
                        return False
                    result = subprocess.run(vcs_cmd, capture_output=True, text=True)  # noqa: PLW1510
                    if result.returncode or not result.stdout.strip():
                        return False
                    latest_commit_id, *_ = result.stdout.split()
                    return commit_id == latest_commit_id

                return False

        return True

    def __getitem__(self, item: str) -> Distribution | None:
        item = self.__canonical_regex.sub("-", item).lower()
        possible_distribution = self.__distributions.get(item)
        if possible_distribution is not None:
            return possible_distribution

        if self.__search_exhausted:
            return None

        for distribution in self.__resolver:
            name = distribution.metadata["Name"]
            if name is None:
                continue

            name = self.__canonical_regex.sub("-", name).lower()
            self.__distributions[name] = distribution
            if name == item:
                return distribution

        self.__search_exhausted = True

        return None


def dependencies_in_sync(
    dependencies: list[Dependency], sys_path: list[str] | None = None, environment: dict[str, str] | None = None
) -> bool:  # no cov
    # This function is unused and only temporarily exists for plugin backwards compatibility.
    distributions = InstalledDistributions(sys_path=sys_path, environment=environment)
    return distributions.dependencies_in_sync(dependencies)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from hatch.env.plugin.interface import EnvironmentInterface


class ExecutionContext:
    def __init__(
        self,
        environment: EnvironmentInterface,
        *,
        shell_commands: list[str] | None = None,
        env_vars: dict[str, str] | None = None,
        force_continue: bool = False,
        show_code_on_error: bool = False,
        hide_commands: bool = False,
        source: str = "cmd",
    ) -> None:
        self.env = environment
        self.shell_commands: list[str] = shell_commands or []
        self.env_vars: dict[str, str] = env_vars or {}
        self.force_continue = force_continue
        self.show_code_on_error = show_code_on_error
        self.hide_commands = hide_commands
        self.source = source

    def add_shell_command(self, command: str | list[str]) -> None:
        self.shell_commands.append(command if isinstance(command, str) else self.env.join_command_args(command))


def parse_matrix_variables(specs: tuple[str, ...]) -> dict[str, set[str]]:
    variables: dict[str, set[str]] = {}
    for spec in specs:
        variable, _, values = spec.partition("=")
        if variable == "py":
            variable = "python"

        if variable in variables:
            raise ValueError(variable)

        variables[variable] = set(values.split(",")) if values else set()

    return variables


def select_environments(
    environments: dict[str, dict[str, Any]],
    included_variables: dict[str, set[str]],
    excluded_variables: dict[str, set[str]],
):
    selected_environments = []
    for env_name, variables in environments.items():
        exclude = False
        for excluded_variable, excluded_values in excluded_variables.items():
            if excluded_variable not in variables:
                continue

            value = variables[excluded_variable]
            if not excluded_values or value in excluded_values:
                exclude = True
                break

        if exclude:
            continue

        for included_variable, included_values in included_variables.items():
            if included_variable not in variables:
                exclude = True
                break

            value = variables[included_variable]
            if included_values and value not in included_values:
                exclude = True
                break

        if not exclude:
            selected_environments.append(env_name)

    return selected_environments



"""Top-level ``hatch lock`` — forwards to the same workflow as ``hatch dep lock``."""

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from hatch.cli.env.lock import dependency_lock_click_options, run_dep_lock

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(
    name="lock",
    short_help="Generate a lockfile for the active environment (alias for `hatch dep lock`)",
)
@dependency_lock_click_options
@click.pass_obj
def lock_command(
    app: Application,
    *,
    upgrade: bool,
    upgrade_package: tuple[str, ...],
    export_path: str | None,
    export_all_path: str | None,
    check: bool,
):
    """Same as :command:`hatch dep lock` for the environment selected with ``-e`` / ``HATCH_ENV``."""
    app.ensure_environment_plugin_dependencies()
    run_dep_lock(
        app,
        upgrade=upgrade,
        upgrade_package=upgrade_package,
        export_path=export_path,
        export_all_path=export_all_path,
        check=check,
    )



"""
Resolve and write environment lockfiles (PEP 751 ``pylock.toml``) via pluggable **lockers**.

Lockers are registered with ``hatch_register_locker``; built-ins ``uv`` and ``pip`` delegate to
``uv pip compile`` / ``pip lock``. See ``docs/how-to/environment/lockfiles.md``.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

    from hatch.env.lockers.interface import LockerInterface
    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.project.core import Project
    from hatch.utils.fs import Path


class LockerNotFoundError(Exception):
    def __init__(self, name: str) -> None:
        super().__init__(f"Unknown locker plugin: {name}")
        self.name = name


class LockerUnsupportedError(Exception):
    def __init__(self, name: str, *, detail: str = "") -> None:
        msg = f"Locker `{name}` does not support this environment"
        if detail:
            msg = f"{msg}: {detail}"
        super().__init__(msg)
        self.name = name
        self.detail = detail


@dataclass
class LockGenerationState:
    """Inputs for :meth:`LockerInterface.generate` / :meth:`LockerInterface.in_sync`."""

    dependencies: list[str]
    layered: bool
    lock_extras: tuple[str, ...]
    lock_groups: tuple[str, ...]


def environment_has_lock_inputs(environment: EnvironmentInterface) -> bool:
    """Whether the environment has anything to pass to the lock resolver."""
    if environment.environment_dependencies or environment.additional_dependencies:
        return True
    if environment.features or environment.dependency_groups:
        return (environment.root / "pyproject.toml").is_file()
    if not environment.skip_install and (environment.root / "pyproject.toml").is_file():
        return True
    return bool(environment.dependencies)


def _lock_environment_only_strings(environment: EnvironmentInterface) -> list[str]:
    lines = list(environment.environment_dependencies)
    lines.extend(map(str, environment.additional_dependencies))
    return lines


def _uv_should_use_layered_lock(
    environment: EnvironmentInterface,
    lock_extras: tuple[str, ...],
    lock_groups: tuple[str, ...],
) -> bool:
    if not (environment.root / "pyproject.toml").is_file():
        return False
    if lock_extras or lock_groups:
        return True
    return not environment.skip_install


def prepare_lock_generation_state(
    environment: EnvironmentInterface,
    *,
    deps_override: list[str] | None = None,
    lock_extras: tuple[str, ...] | None = None,
    lock_groups: tuple[str, ...] | None = None,
) -> LockGenerationState | None:
    """
    Compute PEP 508 dependency lines and layered flags for the active locker.

    Returns ``None`` when there is nothing to lock (same conditions as skipping generation).
    """
    from hatch.env.virtual import VirtualEnvironment

    extras = lock_extras if lock_extras is not None else tuple(environment.features)
    groups = lock_groups if lock_groups is not None else tuple(environment.dependency_groups)

    uv = isinstance(environment, VirtualEnvironment) and environment.use_uv
    layered = uv and _uv_should_use_layered_lock(environment, extras, groups)

    if layered:
        direct_deps = deps_override if deps_override is not None else _lock_environment_only_strings(environment)
    else:
        direct_deps = deps_override if deps_override is not None else list(environment.dependencies)
        extras = ()
        groups = ()

    pyproject = environment.root / "pyproject.toml"
    if layered and not direct_deps and not pyproject.is_file():
        return None

    if not direct_deps and not (layered and pyproject.is_file()):
        return None

    return LockGenerationState(
        dependencies=list(direct_deps),
        layered=layered,
        lock_extras=extras,
        lock_groups=groups,
    )


def get_locker_plugin_class(project: Project, environment: EnvironmentInterface) -> type[LockerInterface]:
    name: str | None = environment.config.get("locker")
    if name is None:
        name = project.config.locker
    if name is None:
        from hatch.env.virtual import VirtualEnvironment

        name = "uv" if isinstance(environment, VirtualEnvironment) and environment.use_uv else "pip"

    cls = project.plugin_manager.locker.get(name)
    if cls is None:
        raise LockerNotFoundError(name)
    if not cls.supports(environment):
        raise LockerUnsupportedError(name)

    return cls


def generate_lockfile(
    environment: EnvironmentInterface,
    output_path: Path,
    *,
    upgrade: bool = False,
    upgrade_packages: tuple[str, ...] = (),
    deps_override: list[str] | None = None,
    lock_extras: tuple[str, ...] | None = None,
    lock_groups: tuple[str, ...] | None = None,
) -> None:
    state = prepare_lock_generation_state(
        environment,
        deps_override=deps_override,
        lock_extras=lock_extras,
        lock_groups=lock_groups,
    )
    if state is None:
        return

    locker_cls = get_locker_plugin_class(environment.app.project, environment)
    locker_cls.generate(
        environment,
        state.dependencies,
        output_path,
        upgrade=upgrade,
        upgrade_packages=upgrade_packages,
        layered=state.layered,
        lock_extras=state.lock_extras,
        lock_groups=state.lock_groups,
    )


def lockfile_in_sync(
    environment: EnvironmentInterface,
    output_path: Path,
    *,
    upgrade: bool = False,
    upgrade_packages: tuple[str, ...] = (),
    deps_override: list[str] | None = None,
    lock_extras: tuple[str, ...] | None = None,
    lock_groups: tuple[str, ...] | None = None,
) -> bool:
    """Whether ``output_path`` matches a fresh resolution (see :meth:`LockerInterface.in_sync`)."""
    state = prepare_lock_generation_state(
        environment,
        deps_override=deps_override,
        lock_extras=lock_extras,
        lock_groups=lock_groups,
    )
    if state is None:
        return output_path.is_file()

    locker_cls = get_locker_plugin_class(environment.app.project, environment)
    return locker_cls.in_sync(
        environment,
        state.dependencies,
        output_path,
        upgrade=upgrade,
        upgrade_packages=upgrade_packages,
        layered=state.layered,
        lock_extras=state.lock_extras,
        lock_groups=state.lock_groups,
    )


verify_lockfile = lockfile_in_sync


def apply_lock_with_locker(environment: EnvironmentInterface, lock_path: Path) -> None:
    """Apply the lockfile at ``lock_path`` using the configured locker."""
    locker_cls = get_locker_plugin_class(environment.app.project, environment)
    locker_cls.apply_lock(environment, lock_path)


def apply_lockfile_to_environment(environment: EnvironmentInterface, lock_path: Path | None = None) -> None:
    """Same as :func:`apply_lock_with_locker` using ``lock_path`` or :func:`resolve_lockfile_path`."""
    path = resolve_lockfile_path(environment) if lock_path is None else lock_path
    apply_lock_with_locker(environment, path)


def merge_environment_lock_inputs(
    project: Project,
    env_names: list[str],
    abort: Callable[[str], None],
) -> tuple[EnvironmentInterface, list[str] | None, tuple[str, ...] | None, tuple[str, ...] | None, str]:
    """
    Merge dependency lines for environments that share one lockfile (matrix / ``lock-filename``).

    Returns ``(representative_environment, deps_override, lock_extras, lock_groups, display_label)``.
    Overrides are ``None`` when only a single environment is involved (use the env's own config).
    """
    from hatch.env.virtual import VirtualEnvironment

    if len(env_names) == 1:
        environment = project.get_environment(env_names[0])
        return environment, None, None, None, env_names[0]

    seen_full: set[str] = set()
    merged_full: list[str] = []
    seen_env_only: set[str] = set()
    merged_env_only: list[str] = []
    python_versions: set[str] = set()
    extras_union: set[str] = set()
    groups_union: set[str] = set()
    for env in env_names:
        env_obj = project.get_environment(env)
        for dep in env_obj.dependencies:
            if dep not in seen_full:
                seen_full.add(dep)
                merged_full.append(dep)
        for dep in env_obj.environment_dependencies:
            if dep not in seen_env_only:
                seen_env_only.add(dep)
                merged_env_only.append(dep)
        for dep in env_obj.additional_dependencies:
            dep_s = str(dep)
            if dep_s not in seen_env_only:
                seen_env_only.add(dep_s)
                merged_env_only.append(dep_s)
        extras_union.update(env_obj.features)
        groups_union.update(env_obj.dependency_groups)
        python_version = env_obj.config.get("python", "")
        python_versions.add(python_version)

    if len(python_versions) > 1:
        versions_str = ", ".join(sorted(v for v in python_versions if v) or ["(default)"])
        abort(
            f"Environments sharing this lockfile target different Python versions ({versions_str}). "
            f"A single lockfile cannot be valid across different Python versions. "
            f"Use distinct `lock-filename` values for each Python version."
        )

    environment = project.get_environment(env_names[0])
    display_name = ", ".join(env_names)
    use_layered_merge = (
        isinstance(environment, VirtualEnvironment)
        and environment.use_uv
        and (environment.root / "pyproject.toml").is_file()
        and (extras_union or groups_union or any(not project.get_environment(e).skip_install for e in env_names))
    )
    if use_layered_merge:
        merged_deps = merged_env_only
        merged_extras = tuple(sorted(extras_union))
        merged_groups = tuple(sorted(groups_union))
    else:
        merged_deps = merged_full
        merged_extras = ()
        merged_groups = ()

    return environment, merged_deps, merged_extras, merged_groups, display_name


def resolve_lockfile_path(environment: EnvironmentInterface) -> Path:
    lock_filename = environment.config.get("lock-filename")
    if lock_filename:
        return environment.root / lock_filename

    if environment.name == "default":
        return environment.root / "pylock.toml"

    # PEP 751 only allows one dot in the filename: pylock.<name>.toml
    safe_name = environment.name.replace(".", "-")
    return environment.root / f"pylock.{safe_name}.toml"



from __future__ import annotations

import json
from functools import cache
from typing import TYPE_CHECKING, Any, Literal

from hatch.utils.fs import Path
from hatch.utils.runner import ExecutionContext

if TYPE_CHECKING:
    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.project.core import Project


class BuildFrontend:
    def __init__(self, project: Project, env: EnvironmentInterface) -> None:
        self.__project = project
        self.__env = env
        self.__scripts = StandardBuildFrontendScripts(self.__project, self.__env)
        self.__hatch = HatchBuildFrontend(self.__project, self.__env)

    @property
    def scripts(self) -> StandardBuildFrontendScripts:
        return self.__scripts

    @property
    def hatch(self) -> HatchBuildFrontend:
        return self.__hatch

    def build_sdist(self, directory: Path) -> Path:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join("output")
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.build_sdist(project_root=self.__env.project_root, output_dir=output_context.env_path)

            script_context = fs_context.join("build_sdist.py")
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(["python", "-u", script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / "output.json"
            output = json.loads(output_path.read_text())

            work_dir = output_context.local_path / "work"
            artifact_path = Path(work_dir / output["return_val"])
            artifact_path.move(directory)
            return directory / artifact_path.name

    def build_wheel(self, directory: Path) -> Path:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join("output")
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.build_wheel(project_root=self.__env.project_root, output_dir=output_context.env_path)

            script_context = fs_context.join("build_wheel.py")
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(["python", "-u", script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / "output.json"
            output = json.loads(output_path.read_text())

            work_dir = output_context.local_path / "work"
            artifact_path = Path(work_dir / output["return_val"])
            artifact_path.move(directory)
            return directory / artifact_path.name

    def get_requires(self, build: Literal["sdist", "wheel", "editable"]) -> list[str]:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join("output")
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.get_requires(
                project_root=self.__env.project_root, output_dir=output_context.env_path, build=build
            )

            script_context = fs_context.join(f"get_requires_{build}.py")
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(["python", "-u", script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / "output.json"
            output = json.loads(output_path.read_text())
            return output["return_val"]

    def get_core_metadata(self, *, editable: bool = False) -> dict[str, Any]:
        from hatchling.metadata.spec import project_metadata_from_core_metadata

        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join("output")
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.prepare_metadata(
                project_root=self.__env.project_root, output_dir=output_context.env_path, editable=editable
            )

            script_context = fs_context.join("get_core_metadata.py")
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(["python", "-u", script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / "output.json"
            output = json.loads(output_path.read_text())

            work_dir = output_context.local_path / "work"
            metadata_file = Path(work_dir) / output["return_val"] / "METADATA"
            return project_metadata_from_core_metadata(metadata_file.read_text())


class HatchBuildFrontend:
    def __init__(self, project: Project, env: EnvironmentInterface) -> None:
        self.__project = project
        self.__env = env
        self.__scripts = HatchBuildFrontendScripts(self.__project, self.__env)

    @property
    def scripts(self) -> HatchBuildFrontendScripts:
        return self.__scripts

    def get_build_deps(self, targets: list[str]) -> list[str]:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join("output")
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.get_build_deps(
                project_root=self.__env.project_root, output_dir=output_context.env_path, targets=targets
            )

            script_context = fs_context.join(f"get_build_deps_{'_'.join(targets)}.py")
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(["python", "-u", script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / "output.json"
            output: list[str] = json.loads(output_path.read_text())
            return output

    def get_core_metadata(self) -> dict[str, Any]:
        with self.__env.fs_context() as fs_context:
            output_context = fs_context.join("output")
            output_context.local_path.ensure_dir_exists()
            script = self.scripts.get_core_metadata(
                project_root=self.__env.project_root, output_dir=output_context.env_path
            )

            script_context = fs_context.join("get_core_metadata.py")
            script_context.local_path.parent.ensure_dir_exists()
            script_context.local_path.write_text(script)
            script_context.sync_env()

            context = ExecutionContext(self.__env)
            context.add_shell_command(["python", "-u", script_context.env_path])
            self.__env.app.execute_context(context)
            output_context.sync_local()

            output_path = output_context.local_path / "output.json"
            output: dict[str, Any] = json.loads(output_path.read_text())
            return output

    def get_required_build_deps(self, targets: list[str]) -> list[str]:
        target_dependencies: list[str] = []
        hooks: set[str] = set()
        for target in targets:
            target_config = self.__project.config.build.target(target)
            target_dependencies.extend(target_config.dependencies)
            hooks.update(target_config.hook_config)

        # Remove any build hooks that are known to not define any dependencies dynamically
        hooks.difference_update((
            # Built-in
            "version",
            # Popular third-party
            "vcs",
        ))

        if hooks:
            return self.get_build_deps(targets)

        return target_dependencies


class BuildFrontendScripts:
    def __init__(self, project: Project, env: EnvironmentInterface) -> None:
        self._project = project
        self._env = env

    @staticmethod
    def inject_data(script: str, data: dict[str, Any]) -> str:
        # All scripts have a constant dictionary on top
        return script.replace("{}", repr(data), 1)


class StandardBuildFrontendScripts(BuildFrontendScripts):
    def get_runner_script(
        self,
        *,
        project_root: str,
        output_dir: str,
        hook: str,
        kwargs: dict[str, Any],
    ) -> str:
        return self.inject_data(
            runner_script(),
            {
                "project_root": project_root,
                "output_dir": output_dir,
                "hook": hook,
                "kwargs": kwargs,
                "backend": self._project.metadata.build.build_backend,
                "backend_path": self._env.pathsep.join(self._project.metadata.build.backend_path),
                "hook_caller_script": hook_caller_script(),
            },
        )

    def get_requires(
        self,
        *,
        project_root: str,
        output_dir: str,
        build: Literal["sdist", "wheel", "editable"],
    ) -> str:
        return self.get_runner_script(
            project_root=project_root,
            output_dir=output_dir,
            hook=f"get_requires_for_build_{build}",
            kwargs={"config_settings": None},
        )

    def prepare_metadata(self, *, output_dir: str, project_root: str, editable: bool = False) -> str:
        return self.get_runner_script(
            project_root=project_root,
            output_dir=output_dir,
            hook="prepare_metadata_for_build_editable" if editable else "prepare_metadata_for_build_wheel",
            kwargs={"work_dir": "metadata_directory", "config_settings": None, "_allow_fallback": True},
        )

    def build_wheel(self, *, output_dir: str, project_root: str, editable: bool = False) -> str:
        return self.get_runner_script(
            project_root=project_root,
            output_dir=output_dir,
            hook="build_editable" if editable else "build_wheel",
            kwargs={"work_dir": "wheel_directory", "config_settings": None, "metadata_directory": None},
        )

    def build_sdist(self, *, output_dir: str, project_root: str) -> str:
        return self.get_runner_script(
            project_root=project_root,
            output_dir=output_dir,
            hook="build_sdist",
            kwargs={"work_dir": "sdist_directory", "config_settings": None},
        )


class HatchBuildFrontendScripts(BuildFrontendScripts):
    def get_build_deps(self, *, output_dir: str, project_root: str, targets: list[str]) -> str:
        return self.inject_data(
            hatch_build_deps_script(),
            {
                "project_root": project_root,
                "output_dir": output_dir,
                "targets": targets,
            },
        )

    def get_core_metadata(self, *, output_dir: str, project_root: str) -> str:
        return self.inject_data(
            hatch_core_metadata_script(),
            {
                "project_root": project_root,
                "output_dir": output_dir,
            },
        )


@cache
def hook_caller_script() -> str:
    from importlib.resources import files

    script = files("pyproject_hooks._in_process") / "_in_process.py"
    return script.read_text(encoding="utf-8")


@cache
def runner_script() -> str:
    from importlib.resources import files

    script = files("hatch.project.frontend.scripts") / "standard.py"
    return script.read_text(encoding="utf-8")


@cache
def hatch_build_deps_script() -> str:
    from importlib.resources import files

    script = files("hatch.project.frontend.scripts") / "build_deps.py"
    return script.read_text(encoding="utf-8")


@cache
def hatch_core_metadata_script() -> str:
    from importlib.resources import files

    script = files("hatch.project.frontend.scripts") / "core_metadata.py"
    return script.read_text(encoding="utf-8")



import os
from tempfile import TemporaryDirectory

from hatch.env.utils import add_verbosity_flag
from hatch.utils.env import PythonInfo
from hatch.utils.fs import Path
from hatch.venv.utils import get_random_venv_name


class VirtualEnv:
    IGNORED_ENV_VARS = ("__PYVENV_LAUNCHER__", "PYTHONHOME")

    def __init__(self, directory, platform, verbosity=0):
        self.directory = directory
        self.platform = platform
        self.verbosity = verbosity
        self.python_info = PythonInfo(platform)

        self._env_vars_to_restore = {}
        self._executables_directory = None

    def activate(self):
        self._env_vars_to_restore["VIRTUAL_ENV"] = os.environ.pop("VIRTUAL_ENV", None)
        os.environ["VIRTUAL_ENV"] = str(self.directory)

        old_path = os.environ.pop("PATH", None)
        self._env_vars_to_restore["PATH"] = old_path
        if old_path is None:
            os.environ["PATH"] = f"{self.executables_directory}{os.pathsep}{os.defpath}"
        else:
            os.environ["PATH"] = f"{self.executables_directory}{os.pathsep}{old_path}"

        for env_var in self.IGNORED_ENV_VARS:
            self._env_vars_to_restore[env_var] = os.environ.pop(env_var, None)

    def deactivate(self):
        for env_var, value in self._env_vars_to_restore.items():
            if value is None:
                os.environ.pop(env_var, None)
            else:
                os.environ[env_var] = value

        self._env_vars_to_restore.clear()

    def create(self, python, *, allow_system_packages=False):
        # WARNING: extremely slow import
        from virtualenv import cli_run

        self.directory.ensure_parent_dir_exists()

        command = [str(self.directory), "--no-download", "--no-periodic-update", "--python", python]

        if allow_system_packages:
            command.append("--system-site-packages")

        # Decrease verbosity since the virtualenv CLI defaults to something like +2 verbosity
        add_verbosity_flag(command, self.verbosity, adjustment=-1)

        cli_run(command)

    def remove(self):
        self.directory.remove()

    def exists(self):
        return self.directory.is_dir()

    @property
    def executables_directory(self):
        if self._executables_directory is None:
            exe_dir = self.directory / ("Scripts" if self.platform.windows else "bin")
            if exe_dir.is_dir():
                self._executables_directory = exe_dir
            # PyPy
            elif self.platform.windows:
                exe_dir = self.directory / "bin"
                if exe_dir.is_dir():
                    self._executables_directory = exe_dir
                else:
                    msg = f"Unable to locate executables directory within: {self.directory}"
                    raise OSError(msg)
            # Debian
            elif (self.directory / "local").is_dir():  # no cov
                exe_dir = self.directory / "local" / "bin"
                if exe_dir.is_dir():
                    self._executables_directory = exe_dir
                else:
                    msg = f"Unable to locate executables directory within: {self.directory}"
                    raise OSError(msg)
            else:
                msg = f"Unable to locate executables directory within: {self.directory}"
                raise OSError(msg)

        return self._executables_directory

    @property
    def environment(self):
        return self.python_info.environment

    @property
    def sys_path(self):
        return self.python_info.sys_path

    def __enter__(self):
        self.activate()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.deactivate()


class TempVirtualEnv(VirtualEnv):
    def __init__(self, parent_python, platform, verbosity=0):
        self.parent_python = parent_python
        self.parent_dir = TemporaryDirectory()
        directory = Path(self.parent_dir.name).resolve() / get_random_venv_name()

        super().__init__(directory, platform, verbosity)

    def remove(self):
        super().remove()
        self.parent_dir.cleanup()

    def __enter__(self):
        self.create(self.parent_python)
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        self.remove()


class UVVirtualEnv(VirtualEnv):
    def create(self, python, *, allow_system_packages=False):
        command = [os.environ.get("HATCH_UV", "uv"), "venv", str(self.directory), "--python", python]
        if allow_system_packages:
            command.append("--system-site-packages")

        add_verbosity_flag(command, self.verbosity, adjustment=-1)
        self.platform.run_command(command)


class TempUVVirtualEnv(TempVirtualEnv, UVVirtualEnv): ...



from __future__ import annotations

from typing import Any

from hatch.env.utils import ensure_valid_environment


def get_internal_env_config() -> dict[str, Any]:
    from hatch.env.internal import build, static_analysis, test, type_check, uv

    internal_config = {}
    for env_name, env_config in (
        ("hatch-build", build.get_default_config()),
        ("hatch-check-code", static_analysis.get_check_code_config()),
        ("hatch-check-fmt", static_analysis.get_check_fmt_config()),
        ("hatch-check-types", type_check.get_default_config()),
        ("hatch-static-analysis", static_analysis.get_default_config()),
        ("hatch-test", test.get_default_config()),
        ("hatch-uv", uv.get_default_config()),
    ):
        env_config["template"] = env_name
        ensure_valid_environment(env_config)
        internal_config[env_name] = env_config

    return internal_config


def is_isolated_environment(env_name: str, config: dict[str, Any]) -> bool:
    # Provide super isolation and immunity to project-level environment removal only when the environment:
    #
    # 1. Is not used for builds
    # 2. Does not require the project being installed
    # 3. The default configuration is used
    #
    # For example, the environment for static analysis depends only on Ruff at a specific default
    # version. This environment does not require the project and can be reused by every project to
    # improve responsiveness. However, if the user for some reason chooses to override the dependencies
    # to use a different version of Ruff, then the project would get its own environment.
    return (
        not config.get("builder", False)
        and config.get("skip-install", False)
        and is_default_environment(env_name, config)
    )


def is_default_environment(env_name: str, config: dict[str, Any]) -> bool:
    # Standalone environment
    internal_config = get_internal_env_config().get(env_name)
    if not internal_config:
        # Environment generated from matrix
        internal_config = get_internal_env_config().get(env_name.split(".")[0])
        if not internal_config:
            return False

    # Only consider things that would modify the actual installation, other options like extra scripts don't matter
    for key in ("dependencies", "extra-dependencies", "features"):
        if config.get(key) != internal_config.get(key):
            return False

    return True



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


def ensure_path_public(path: str, shells: list[str]) -> bool:
    import userpath

    if userpath.in_current_path(path) or userpath.in_new_path(path, shells):
        return True

    userpath.append(path, shells=shells)
    return False


@click.command(short_help="Install Python distributions")
@click.argument("names", required=True, nargs=-1)
@click.option("--private", is_flag=True, help="Do not add distributions to the user PATH")
@click.option("--update", "-u", is_flag=True, help="Update existing installations")
@click.option(
    "--dir", "-d", "directory", help="The directory in which to install distributions, overriding configuration"
)
@click.pass_obj
def install(app: Application, *, names: tuple[str, ...], private: bool, update: bool, directory: str | None):
    """
    Install Python distributions.

    You may select `all` to install all compatible distributions:

    \b
    ```
    hatch python install all
    ```

    You can set custom sources for distributions by setting the `HATCH_PYTHON_SOURCE_<NAME>` environment variable
    where `<NAME>` is the uppercased version of the distribution name with periods replaced by underscores e.g.
    `HATCH_PYTHON_SOURCE_PYPY3_10`.
    """
    from hatch.errors import PythonDistributionResolutionError, PythonDistributionUnknownError
    from hatch.python.distributions import ORDERED_DISTRIBUTIONS
    from hatch.python.resolve import get_distribution

    shells = []
    if not private and not app.platform.windows:
        shell_name, _ = app.shell_data
        shells.append(shell_name)

    manager = app.get_python_manager(directory)
    installed = manager.get_installed()
    selection = ORDERED_DISTRIBUTIONS if "all" in names else names
    unknown = []
    compatible = []
    incompatible = []
    for name in selection:
        if name in installed:
            compatible.append(name)
            continue

        try:
            get_distribution(name)
        except PythonDistributionUnknownError:
            unknown.append(name)
        except PythonDistributionResolutionError:
            incompatible.append(name)
        else:
            compatible.append(name)

    if unknown:
        app.abort(f"Unknown distributions: {', '.join(unknown)}")
    elif incompatible and (not compatible or "all" not in names):
        app.abort(f"Incompatible distributions: {', '.join(incompatible)}")

    directories_made_public = []
    for name in compatible:
        needs_update = False
        if name in installed:
            installed_dist = installed[name]
            needs_update = installed_dist.needs_update()
            if not needs_update:
                app.display_warning(f"The latest version is already installed: {installed_dist.version}")
                continue

            if not (update or app.confirm(f"Update {name}?")):
                app.abort(f"Distribution is already installed: {name}")

        with app.status(f"{'Updating' if needs_update else 'Installing'} {name}"):
            dist = manager.install(name)
            if not private:
                python_directory = str(dist.python_path.parent)
                if not ensure_path_public(python_directory, shells=shells):
                    directories_made_public.append(python_directory)

        app.display_success(f"{'Updated' if needs_update else 'Installed'} {name} @ {dist.path}")

    if directories_made_public:
        multiple = len(directories_made_public) > 1
        app.display(
            f"\nThe following director{'ies' if multiple else 'y'} ha{'ve' if multiple else 's'} "
            f"been added to your PATH (pending a shell restart):\n"
        )
        for public_directory in directories_made_public:
            app.display(public_directory)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING

import hyperlink

from hatch._version import __version__

if TYPE_CHECKING:
    import httpx2

    from hatch.utils.fs import Path


class IndexURLs:
    def __init__(self, repo: str):
        self.repo = hyperlink.parse(repo).normalize()

        # PyPI
        if self.repo.host.endswith("pypi.org"):  # no cov
            repo_url = self.repo.replace(host="pypi.org") if self.repo.host == "upload.pypi.org" else self.repo

            self.simple = repo_url.click("/simple/")
            self.project = repo_url.click("/project/")

        # Assume devpi
        else:
            self.simple = self.repo.child("+simple", "")
            self.project = self.repo


class PackageIndex:
    def __init__(self, repo: str, *, user="", auth="", ca_cert=None, client_cert=None, client_key=None):
        self.urls = IndexURLs(repo)
        self.repo = str(self.urls.repo)
        self.user = user
        self.auth = auth

        self.__cert = None
        if client_cert:
            self.__cert = client_cert
            if client_key:
                self.__cert = (client_cert, client_key)

        self.__verify = True
        if ca_cert:
            self.__verify = ca_cert

    @cached_property
    def client(self) -> httpx2.Client:
        import httpx2

        from hatch.utils.linehaul import get_linehaul_component
        from hatch.utils.network import DEFAULT_TIMEOUT

        user_agent = f"Hatch/{__version__} {get_linehaul_component()} HTTPX2/{httpx2.__version__}"
        return httpx2.Client(
            headers={"User-Agent": user_agent},
            transport=httpx2.HTTPTransport(retries=3, verify=self.__verify, cert=self.__cert),
            timeout=DEFAULT_TIMEOUT,
        )

    def upload_artifact(self, artifact: Path, data: dict):
        import hashlib
        import io

        data[":action"] = "file_upload"
        data["protocol_version"] = "1"

        with artifact.open("rb") as f:
            # https://github.com/pypa/warehouse/blob/7fc3ce5bd7ecc93ef54c1652787fb5e7757fe6f2/tests/unit/packaging/test_tasks.py#L189-L191
            md5_hash = hashlib.md5()  # noqa: S324
            sha256_hash = hashlib.sha256()
            blake2_256_hash = hashlib.blake2b(digest_size=32)

            while True:
                chunk = f.read(io.DEFAULT_BUFFER_SIZE)
                if not chunk:
                    break

                md5_hash.update(chunk)
                sha256_hash.update(chunk)
                blake2_256_hash.update(chunk)

            data["md5_digest"] = md5_hash.hexdigest()
            data["sha256_digest"] = sha256_hash.hexdigest()
            data["blake2_256_digest"] = blake2_256_hash.hexdigest()

            f.seek(0)

            response = self.client.post(
                self.repo,
                data=data,
                files={"content": (artifact.name, f, "application/octet-stream")},
                auth=(self.user, self.auth),
            )
            response.raise_for_status()

    def get_simple_api(self, project: str) -> httpx2.Response:
        return self.client.get(
            str(self.urls.simple.child(project, "")),
            headers={"Cache-Control": "no-cache"},
            auth=(self.user, self.auth),
        )
